import pathlib

import pytest

from rbx.box import header
from rbx.box.testing import testing_package


class TestHeader:
    """Tests for the header.py module."""

    def test_generate_header_with_no_variables(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test generate_header with a package that has no variables."""
        # Generate header with no variables
        header.generate_header()

        # Check that rbx.h file was created
        rbx_header_path = pathlib.Path('rbx.h')
        assert rbx_header_path.exists()

        # Read the generated header
        generated_content = rbx_header_path.read_text()

        # Check that the header contains the expected structure
        assert (
            'std::optional<std::string> getStringVar(std::string name)'
            in generated_content
        )
        assert 'std::optional<int64_t> getIntVar(std::string name)' in generated_content
        assert 'std::optional<float> getFloatVar(std::string name)' in generated_content
        assert 'std::optional<bool> getBoolVar(std::string name)' in generated_content

        # Since there are no variables, the functions should only return nullopt
        assert 'return std::nullopt;' in generated_content

    def test_generate_header_with_string_variables(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test generate_header with string variables."""
        # Set some string variables
        testing_pkg.set_vars(
            {
                'str_var1': 'hello',
                'str_var2': 'world with spaces',
                'str_var3': 'special"chars',
            }
        )

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Check that string variables are properly escaped and included
        assert (
            'if (name == "str_var1") {\n    return "hello";\n  }' in generated_content
        )
        assert (
            'if (name == "str_var2") {\n    return "world with spaces";\n  }'
            in generated_content
        )
        assert (
            'if (name == "str_var3") {\n    return "special\\"chars";\n  }'
            in generated_content
        )

    def test_generate_header_with_int_variables(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test generate_header with integer variables."""
        testing_pkg.set_vars(
            {
                'int_var1': 42,
                'int_var2': -100,
                'int_var3': 0,
            }
        )

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Check that integer variables are properly included
        assert (
            'if (name == "int_var1") {\n    return static_cast<int64_t>(42);\n  }'
            in generated_content
        )
        assert (
            'if (name == "int_var2") {\n    return static_cast<int64_t>(-100);\n  }'
            in generated_content
        )
        assert (
            'if (name == "int_var3") {\n    return static_cast<int64_t>(0);\n  }'
            in generated_content
        )

    def test_generate_header_with_float_variables(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test generate_header with float variables."""
        testing_pkg.set_vars(
            {
                'float_var1': 3.14,
                'float_var2': -2.5,
                'float_var3': 0.0,
            }
        )

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Check that float variables are properly included
        assert 'if (name == "float_var1") {\n    return 3.14;\n  }' in generated_content
        assert 'if (name == "float_var2") {\n    return -2.5;\n  }' in generated_content
        assert 'if (name == "float_var3") {\n    return 0.0;\n  }' in generated_content

    def test_generate_header_with_bool_variables(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test generate_header with boolean variables."""
        testing_pkg.set_vars(
            {
                'bool_var1': True,
                'bool_var2': False,
            }
        )

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Check that boolean variables are properly included
        assert 'if (name == "bool_var1") {\n    return true;\n  }' in generated_content
        assert 'if (name == "bool_var2") {\n    return false;\n  }' in generated_content

    def test_generate_header_with_mixed_variables(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test generate_header with variables of different types."""
        testing_pkg.set_vars(
            {
                'str_var': 'test',
                'int_var': 123,
                'float_var': 45.67,
                'bool_var': True,
            }
        )

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Check that all variable types are included in their respective functions
        assert 'if (name == "str_var") {\n    return "test";\n  }' in generated_content
        assert (
            'if (name == "int_var") {\n    return static_cast<int64_t>(123);\n  }'
            in generated_content
        )
        assert 'if (name == "float_var") {\n    return 45.67;\n  }' in generated_content
        assert 'if (name == "bool_var") {\n    return true;\n  }' in generated_content

    def test_generate_header_with_override_file(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test generate_header with an override file."""
        override_content = '// This is an override header\n'
        override_path = pathlib.Path('rbx.override.h')
        override_path.write_text(override_content)

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Should use the override content exactly
        assert generated_content == override_content

    def test_get_header_creates_and_returns_path(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test get_header function creates the header and returns the path."""
        result = header.get_header()

        assert isinstance(result, pathlib.Path)
        assert result.name == 'rbx.h'
        assert result.exists()

    def test_get_header_caching(self, testing_pkg: testing_package.TestingPackage):
        """Test that get_header is cached properly."""
        # First call should create the file
        result1 = header.get_header()

        # Modify the file to test caching
        result1.write_text('modified content')

        # Second call should return the same path (cached)
        result2 = header.get_header()

        assert result1 == result2
        assert result2.read_text() == 'modified content'

    def test_check_int_bounds_valid_values(self):
        """Test _check_int_bounds with valid integer values."""
        # These should not raise exceptions
        header.check_int_bounds(0)
        header.check_int_bounds(100)
        header.check_int_bounds(-100)
        header.check_int_bounds(2**63 - 1)  # Max int64_t
        header.check_int_bounds(-(2**63))  # Min int64_t

    def test_check_int_bounds_too_large(self):
        """Test _check_int_bounds with values that are too large."""
        with pytest.raises(
            ValueError, match='too large to fit in a C\\+\\+ 64-bit integer'
        ):
            header.check_int_bounds(2**64)

    def test_check_int_bounds_too_small(self):
        """Test _check_int_bounds with values that are too small."""
        with pytest.raises(
            ValueError, match='too small to fit in a C\\+\\+ 64-bit signed integer'
        ):
            header.check_int_bounds(-(2**63) - 1)

    def test_generate_header_with_large_int_fails(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that generate_header fails with integers that are too large."""
        testing_pkg.set_vars(
            {
                'large_int': 2**64,  # Too large for int64_t
            }
        )

        with pytest.raises(
            ValueError, match='too large to fit in a C\\+\\+ 64-bit integer'
        ):
            header.generate_header()

    def test_generate_header_with_small_int_fails(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that generate_header fails with integers that are too small."""
        testing_pkg.set_vars(
            {
                'small_int': -(2**63) - 1,  # Too small for int64_t
            }
        )

        with pytest.raises(
            ValueError, match='too small to fit in a C\\+\\+ 64-bit signed integer'
        ):
            header.generate_header()

    def test_generate_header_deterministic_order(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that generate_header produces variables in deterministic order."""
        testing_pkg.set_vars(
            {
                'z_var': 'last',
                'a_var': 'first',
                'm_var': 'middle',
            }
        )

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Variables should be in alphabetical order
        a_pos = generated_content.find('if (name == "a_var")')
        m_pos = generated_content.find('if (name == "m_var")')
        z_pos = generated_content.find('if (name == "z_var")')

        assert a_pos < m_pos < z_pos

    def test_string_repr_escaping(self, testing_pkg: testing_package.TestingPackage):
        """Test that string variables with special characters are properly escaped in generated header."""
        # Set variables with various special characters that need escaping
        testing_pkg.set_vars(
            {
                'simple_string': 'simple',
                'string_with_spaces': 'with spaces',
                'string_with_quotes': 'with"quotes',
                'string_with_backslash': 'with\\backslash',
                'string_with_newline': 'with\nnewline',
                'string_with_tab': 'with\ttab',
                'string_with_return': 'with\rreturn',
            }
        )

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Check that all string variables are properly escaped in the generated header
        assert (
            'if (name == "simple_string") {\n    return "simple";\n  }'
            in generated_content
        )
        assert (
            'if (name == "string_with_spaces") {\n    return "with spaces";\n  }'
            in generated_content
        )
        assert (
            'if (name == "string_with_quotes") {\n    return "with\\"quotes";\n  }'
            in generated_content
        )
        assert (
            'if (name == "string_with_backslash") {\n    return "with\\\\backslash";\n  }'
            in generated_content
        )
        assert (
            'if (name == "string_with_newline") {\n    return "with\\nnewline";\n  }'
            in generated_content
        )
        assert (
            'if (name == "string_with_tab") {\n    return "with\\ttab";\n  }'
            in generated_content
        )
        assert (
            'if (name == "string_with_return") {\n    return "with\\rreturn";\n  }'
            in generated_content
        )

    def test_generate_header_bool_as_int_in_int_block(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that bool variables are handled correctly in int context."""
        testing_pkg.set_vars(
            {
                'bool_true': True,
                'bool_false': False,
            }
        )

        header.generate_header()

        rbx_header_path = pathlib.Path('rbx.h')
        generated_content = rbx_header_path.read_text()

        # Booleans should appear in both bool and int blocks
        # In int block, they should be converted to 1/0
        assert (
            'if (name == "bool_true") {\n    return static_cast<int64_t>(1);\n  }'
            in generated_content
        )
        assert (
            'if (name == "bool_false") {\n    return static_cast<int64_t>(0);\n  }'
            in generated_content
        )

        # In bool block, they should be true/false
        assert 'if (name == "bool_true") {\n    return true;\n  }' in generated_content
        assert (
            'if (name == "bool_false") {\n    return false;\n  }' in generated_content
        )



"""
File for defining mkdocs macros.
"""


def define_env(env):
    @env.macro
    def asciinema(id: str, idleness: float = 1, speed: float = 1):
        return f"""<div style="width: 90%; margin: 0 auto;">
<script src="https://asciinema.org/a/{id}.js" id="asciicast-{id}" async="true" data-autoplay data-loop data-idle-time-limit="{idleness}" data-speed="{speed}"></script>
</div>
"""



import pathlib
import sys
from enum import Enum
from typing import Dict, List, Union, get_args, get_origin, get_type_hints

import mkdocs_gen_files
import yaml
from click import Argument, Choice, Command, Group, Option
from typer.main import get_command

# Fix for shadowing 'packaging' library by 'rbx/box/packaging' directory
# when the script directory is in sys.path.
script_dir = str(pathlib.Path(__file__).parent.resolve())
if script_dir in sys.path:
    sys.path.remove(script_dir)

# Import the main app
from rbx.box.cli import app as main_app  # noqa: E402

# Path to custom explanations
EXPLANATIONS_PATH = pathlib.Path('docs/setters/reference/cli_explanations.yaml')


def load_explanations() -> Dict[str, str]:
    if not EXPLANATIONS_PATH.exists():
        return {}
    with open(EXPLANATIONS_PATH, 'r') as f:
        return yaml.safe_load(f) or {}


EXPLANATIONS = load_explanations()


def unwrap_type(tp):
    origin = get_origin(tp)
    if origin is Union:
        args = get_args(tp)
        non_none = [a for a in args if a is not type(None)]
        if len(non_none) == 1:
            return unwrap_type(non_none[0])
        return tp

    # Handle Annotated
    if origin is not None and getattr(origin, '__name__', '') == 'Annotated':
        return unwrap_type(get_args(tp)[0])

    return tp


def get_enum_link(tp) -> Union[str, None]:
    try:
        if isinstance(tp, type) and issubclass(tp, Enum):
            return f'[{tp.__name__}][{tp.__module__}.{tp.__name__}]'
    except Exception:
        pass
    return None


class DocsGenerator:
    def __init__(self):
        self.content = []
        self.seen_enums = set()

    def get_full_command_path(self, lineage: List[str]) -> str:
        return ' '.join(lineage)

    def render_command(self, command: Command, lineage: List[str]) -> str:
        """Renders a single command to Markdown."""

        # Construct normalized path for explanation lookup
        # Normalize each part of lineage: "build, b" -> "build"
        normalized_lineage = [part.split(',')[0].strip() for part in lineage]
        full_path_normalized = ' '.join(normalized_lineage)

        hints = {}
        if command.callback:
            try:
                hints = get_type_hints(command.callback)
            except Exception:
                pass

        # Header
        output = []

        # Fallback to normalized path string if get_command_help isn't updated?
        # Actually get_command_help expects string. We can update it or just call dict directly
        # since we did normalization here.
        # Let's just use EXPLANATIONS directly since we normalized it here correctly.
        custom_help = EXPLANATIONS.get(full_path_normalized)

        if custom_help:
            output.append(custom_help)
            output.append('')
        elif command.help:
            output.append(command.help)
            output.append('')

        # Usage
        output.append('**Usage:**')
        usage_pieces = normalized_lineage[:]  # Copy

        # Collect arguments and options for usage string
        params = command.params
        for param in params:
            if isinstance(param, Argument):
                usage_pieces.append(f'<{param.name.upper()}>')
            elif isinstance(param, Option):
                pass  # We typically don't list all options in the usage line for complex CLIs, but we can if relevant

        output.append(f'```bash\n{" ".join(usage_pieces)} [OPTIONS]\n```')
        output.append('')

        # Arguments
        arguments = [p for p in params if isinstance(p, Argument)]
        if arguments:
            output.append('**Arguments:**')
            output.append('')
            output.append('| Name | Description | Required |')
            output.append('| :--- | :--- | :--- |')
            for arg in arguments:
                desc = arg.help or '-'
                required = 'Yes' if arg.required else 'No'
                output.append(f'| `{arg.name.upper()}` | {desc} | {required} |')
            output.append('')

        # Options
        options = [p for p in params if isinstance(p, Option)]
        if options:
            output.append('| Name | Type | Description | Default |')
            output.append('| :--- | :--- | :--- | :--- |')
            for opt in options:
                opts = ', '.join(f'`{o}`' for o in opt.opts)
                opt_type = (
                    opt.type.name.upper() if hasattr(opt.type, 'name') else 'TEXT'
                )
                desc = opt.help or '-'

                # Type resolution
                param_name = opt.name
                target_type = None
                if param_name:
                    target_type = hints.get(param_name) or hints.get(
                        param_name.replace('-', '_')
                    )

                enum_link = None

                # Special handling for VerificationLevel
                if param_name in ('verification', 'verification_level'):
                    from rbx.box.environment import VerificationLevel

                    self.seen_enums.add(VerificationLevel)
                    enum_link = f'INTEGER of [{VerificationLevel.__name__}][{VerificationLevel.__module__}.{VerificationLevel.__name__}]'
                elif target_type:
                    unwrapped = unwrap_type(target_type)
                    # Register enum if found
                    if isinstance(unwrapped, type) and issubclass(unwrapped, Enum):
                        self.seen_enums.add(unwrapped)
                    enum_link = get_enum_link(unwrapped)

                if enum_link:
                    opt_type = enum_link
                elif isinstance(opt.type, Choice):
                    desc += f' <br> **Choices:** {", ".join(f"`{c}`" for c in opt.type.choices)}'

                default = opt.default
                if callable(default):
                    default = default()

                if default is not None and str(default) != '':
                    default = f'`{default}`'
                else:
                    default = '-'
                output.append(f'| {opts} | {opt_type} | {desc} | {default} |')
            output.append('')

        return '\n'.join(output)

    def process_command(self, command: Command, lineage: List[str]):
        name = lineage[-1]
        # Format name for display: "list, ls" -> "list (ls)"
        parts = [p.strip() for p in name.split(',')]
        primary_name = parts[0]
        aliases = parts[1:]

        display_name = primary_name
        if aliases:
            display_name += f' ({", ".join(aliases)})'

        # Section header
        level = min(len(lineage), 6)  # h1, h2, h3...

        # If it's the root 'rbx', we use h1
        # If 'rbx build', h2
        # etc.

        self.content.append(f'{"#" * level} {display_name}\n')
        self.content.append(self.render_command(command, lineage))
        self.content.append('\n---\n')

        if isinstance(command, Group):
            for sub_name, sub_cmd in command.commands.items():
                if sub_cmd.hidden:
                    continue
                self.process_command(sub_cmd, lineage + [sub_name])

    def render_enums(self):
        if not self.seen_enums:
            return

        self.content.append('# Generic Types\n')
        # Sort by name for stability
        sorted_enums = sorted(self.seen_enums, key=lambda e: e.__name__)
        for enum_cls in sorted_enums:
            self.content.append(f'::: {enum_cls.__module__}.{enum_cls.__name__}')
            self.content.append('    options:')
            self.content.append('      show_root_heading: true')
            self.content.append('      show_root_full_path: false')
            self.content.append('      heading_level: 3')
            self.content.append('      show_labels: false')
            self.content.append('      members_order: source')
            self.content.append('')


def generate():
    # Convert Typer app to Click command
    main_cmd = get_command(main_app)

    # We want to name the root command 'rbx'
    # typer sometimes names it 'main' or similar depending on function name

    generator = DocsGenerator()
    generator.process_command(main_cmd, ['rbx'])
    generator.render_enums()

    # Write to file
    with mkdocs_gen_files.open('setters/reference/cli.md', 'w') as f:
        f.write('\n'.join(generator.content))


generate()



import dataclasses
import json
import pathlib
from typing import Any, Dict, Iterable, Iterator, List, Optional, Set, Tuple

from TexSoup.data import BraceGroup, BracketGroup, TexCmd, TexNode

from rbx.box.statements.texsoup_utils import parse_latex


@dataclasses.dataclass
class MacroDef:
    name: str
    n_args: int
    default: Optional[str]
    body: str
    source_file: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'MacroDef':
        return cls(**data)


class MacroDefinitions:
    def __init__(self) -> None:
        self._defs: Dict[str, MacroDef] = {}

    def add(self, macro_def: MacroDef) -> None:
        self._defs[macro_def.name] = macro_def

    def get(self, name: str) -> Optional[MacroDef]:
        return self._defs.get(name)

    def merge(self, other: 'MacroDefinitions') -> None:
        for name in other:
            macro = other.get(name)
            if macro is not None:
                self.add(macro)

    def __contains__(self, name: str) -> bool:
        return name in self._defs

    def __iter__(self) -> Iterator[str]:
        return iter(self._defs)

    def __len__(self) -> int:
        return len(self._defs)

    def filter(self, to_filter_out: Iterable[str]) -> 'MacroDefinitions':
        to_filter_out = set(to_filter_out)
        new_defs = MacroDefinitions()
        for name, macro in self._defs.items():
            if name not in to_filter_out:
                new_defs.add(macro)
        return new_defs

    def to_json_file(self, path: pathlib.Path) -> None:
        data = [macro.to_dict() for macro in self._defs.values()]
        path.write_text(
            json.dumps(data, ensure_ascii=False, indent=2), encoding='utf-8'
        )

    @classmethod
    def from_json_file(cls, path: pathlib.Path) -> 'MacroDefinitions':
        data = json.loads(path.read_text(encoding='utf-8'))
        defs = cls()
        for entry in data:
            defs.add(MacroDef.from_dict(entry))
        return defs


_NEWCOMMAND_NAMES = {'newcommand', 'newcommand*', 'renewcommand', 'renewcommand*'}


def _extract_name_from_arg(arg) -> Optional[str]:
    """Extract command name (without backslash) from a newcommand's first argument."""
    if isinstance(arg, BraceGroup):
        for child in arg.contents:
            if isinstance(child, TexCmd):
                return child.name
        text = arg.string.strip()
        if text.startswith('\\'):
            return text[1:]
    return None


def _extract_newcommand(
    node: TexNode, source_file: Optional[str] = None
) -> Optional[MacroDef]:
    """Extract a MacroDef from a \\newcommand or \\renewcommand node."""
    args = list(node.args)
    if not args:
        return None

    name = _extract_name_from_arg(args[0])
    if name is None:
        return None

    idx = 1
    n_args = 0
    default = None
    body = ''

    if idx < len(args) and isinstance(args[idx], BracketGroup):
        try:
            n_args = int(args[idx].string)
        except ValueError:
            pass
        idx += 1

    if idx < len(args) and isinstance(args[idx], BracketGroup):
        default = args[idx].string
        idx += 1

    if idx < len(args) and isinstance(args[idx], BraceGroup):
        body = args[idx].string

    return MacroDef(
        name=name,
        n_args=n_args,
        default=default,
        body=body,
        source_file=source_file,
    )


def _extract_def(
    node: TexNode, source_file: Optional[str] = None
) -> Optional[MacroDef]:
    """Extract a MacroDef from a \\def node (0-arg form only)."""
    args = list(node.args)
    if len(args) < 2:
        return None

    first = args[0]
    if isinstance(first, TexCmd):
        name = first.name
    elif isinstance(first, BraceGroup):
        text = first.string.strip()
        if text.startswith('\\'):
            name = text[1:]
        else:
            return None
    else:
        return None

    if isinstance(args[1], BraceGroup):
        body = args[1].string
    else:
        return None

    return MacroDef(
        name=name,
        n_args=0,
        default=None,
        body=body,
        source_file=source_file,
    )


def extract_definitions(
    tex_content: str,
    source_file: Optional[str] = None,
) -> MacroDefinitions:
    """Extract all macro definitions from a TeX string."""
    soup = parse_latex(tex_content)
    defs = MacroDefinitions()

    # Iterate descendants in document order so that later definitions
    # (e.g. \renewcommand) correctly overwrite earlier ones.
    for node in soup.descendants:
        if not isinstance(node, TexNode):
            continue
        name = getattr(node, 'name', None)
        if name in _NEWCOMMAND_NAMES:
            macro = _extract_newcommand(node, source_file)
            if macro is not None:
                defs.add(macro)
        elif name == 'def':
            macro = _extract_def(node, source_file)
            if macro is not None:
                defs.add(macro)

    return defs


def _resolve_input_path(
    name: str, base_dir: pathlib.Path, extensions: tuple = ('.tex',)
) -> Optional[pathlib.Path]:
    """Resolve an \\input/\\include filename to an actual file path."""
    candidate = base_dir / name
    if candidate.is_file():
        return candidate.resolve()
    for ext in extensions:
        with_ext = base_dir / (name + ext)
        if with_ext.is_file():
            return with_ext.resolve()
    return None


def _resolve_package_path(name: str, base_dir: pathlib.Path) -> Optional[pathlib.Path]:
    """Resolve a \\usepackage/\\RequirePackage name to a local .sty file."""
    candidate = base_dir / (name + '.sty')
    if candidate.is_file():
        return candidate.resolve()
    return None


def _collect_recursive(
    tex_path: pathlib.Path,
    base_dir: pathlib.Path,
    visited: Set[pathlib.Path],
) -> MacroDefinitions:
    resolved = tex_path.resolve()
    if resolved in visited:
        return MacroDefinitions()
    visited.add(resolved)

    try:
        content = tex_path.read_text(encoding='utf-8')
    except (OSError, UnicodeDecodeError):
        return MacroDefinitions()

    defs = extract_definitions(content, source_file=str(resolved))

    soup = parse_latex(content)

    for cmd_name in ('input', 'include'):
        for node in soup.find_all(cmd_name):
            args = list(node.args)
            if not args:
                continue
            ref_name = args[0].string.strip()
            ref_path = _resolve_input_path(ref_name, base_dir)
            if ref_path is not None:
                child_defs = _collect_recursive(ref_path, base_dir, visited)
                defs.merge(child_defs)

    for cmd_name in ('usepackage', 'RequirePackage'):
        for node in soup.find_all(cmd_name):
            args = list(node.args)
            if not args:
                continue
            pkg_arg = args[-1].string.strip()
            for pkg_name in pkg_arg.split(','):
                pkg_name = pkg_name.strip()
                if not pkg_name:
                    continue
                pkg_path = _resolve_package_path(pkg_name, base_dir)
                if pkg_path is not None:
                    child_defs = _collect_recursive(pkg_path, base_dir, visited)
                    defs.merge(child_defs)

    return defs


def collect_macro_definitions(
    tex_path: pathlib.Path,
    base_dir: Optional[pathlib.Path] = None,
) -> MacroDefinitions:
    """Collect all macro definitions from a TeX file, recursively visiting dependencies."""
    if base_dir is None:
        base_dir = tex_path.parent
    return _collect_recursive(tex_path, base_dir, set())


def _substitute_body(body: str, args: List[str]) -> str:
    """Replace #1, #2, ... in body with actual argument strings."""
    result = body
    for i, arg_val in enumerate(args, 1):
        result = result.replace(f'#{i}', arg_val)
    return result


def _find_and_expand(tex_content: str, macro_defs: MacroDefinitions) -> Tuple[str, int]:
    """Single pass: find macro usages via TexSoup and expand them via text replacement."""
    try:
        soup = parse_latex(tex_content)
    except Exception:
        return tex_content, 0

    replacements: List[Tuple[int, int, str]] = []

    def visit(node: TexNode) -> None:
        if not isinstance(node, TexNode):
            return

        name = getattr(node, 'name', None)
        if name and name in macro_defs:
            macro = macro_defs.get(name)
            if macro is None:
                return
            pos = getattr(node, 'position', None)
            if pos is None:
                return

            node_args = list(node.args)
            arg_strings: List[str] = []

            if macro.n_args == 0:
                # 0-arg macro: don't consume any TexSoup-parsed arguments.
                # TexSoup may greedily parse following braces as args.
                start = pos
                end = start + len('\\' + name)
            else:
                start = pos
                end = start + len(str(node))

                if macro.default is not None:
                    # First arg is optional (BracketGroup).
                    if node_args and isinstance(node_args[0], BracketGroup):
                        arg_strings.append(node_args[0].string)
                        remaining = node_args[1:]
                    else:
                        arg_strings.append(macro.default)
                        remaining = node_args
                    for arg in remaining:
                        arg_strings.append(arg.string)
                else:
                    for arg in node_args:
                        arg_strings.append(arg.string)

            expanded = _substitute_body(macro.body, arg_strings)
            replacements.append((start, end, expanded))
            return

        for child in node.contents:
            if isinstance(child, TexNode):
                visit(child)

    visit(soup)

    if not replacements:
        return tex_content, 0

    # Apply replacements right-to-left to preserve earlier positions.
    replacements.sort(key=lambda r: r[0], reverse=True)
    result = tex_content
    for start, end, replacement in replacements:
        result = result[:start] + replacement + result[end:]

    return result, len(replacements)


def expand_macros(
    tex_content: str,
    macro_defs: MacroDefinitions,
    max_iterations: int = 10,
) -> str:
    """Expand all macro usages in a TeX string.

    Uses an iterative fixpoint approach: each iteration parses the current text
    with TexSoup, finds macro usages, and expands them via text replacement.
    Repeats until no more substitutions are made or *max_iterations* is reached
    (handles nested macros).
    """
    content = tex_content
    for _ in range(max_iterations):
        content, count = _find_and_expand(content, macro_defs)
        if count == 0:
            break
    return content



import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Literal, Optional, Set

import rich
import rich.progress
import typer

from rbx import console, utils
from rbx.box import download, header, naming, package
from rbx.box.packaging.polygon import polygon_api as api
from rbx.box.packaging.polygon.statement_block_utils import (
    get_processed_statement_blocks,
    process_statements,
)
from rbx.box.packaging.polygon.utils import get_polygon_language_from_code_item
from rbx.box.schema import (
    ExpectedOutcome,
    Generator,
    GeneratorCall,
    Solution,
    TaskType,
    Testcase,
)
from rbx.box.solutions import get_best_interaction_file
from rbx.box.statements.build_statements import get_produced_tikz_pdfs
from rbx.box.statements.builders import (
    StatementSample,
)
from rbx.box.statements.schema import Statement
from rbx.box.statements.statement_utils import get_relative_assets
from rbx.box.testcase_extractors import extract_generation_testcases_from_groups
from rbx.box.testcase_sample_utils import get_statement_samples
from rbx.box.testcase_utils import (
    TestcaseInteractionParsingError,
    get_alternate_interaction_texts,
    parse_interaction,
)

_API_URL = 'https://polygon.codeforces.com/api'

ParamChoices = Literal['statements', 'solutions', 'tests', 'files']

ALL_PARAMS_CHOICES = list(ParamChoices.__args__)
MAX_WORKERS = 4


def _get_polygon_api() -> api.Polygon:
    env = utils.environ()
    return api.Polygon(
        _API_URL,
        env.get('POLYGON_API_KEY', '').strip(),
        env.get('POLYGON_API_SECRET', '').strip(),
    )


def _get_solution_tag(solution: Solution, is_first: bool = False) -> api.SolutionTag:
    if solution.outcome == ExpectedOutcome.ACCEPTED:
        return api.SolutionTag.OK if not is_first else api.SolutionTag.MA
    if solution.outcome == ExpectedOutcome.ACCEPTED_OR_TLE:
        return api.SolutionTag.TO
    if solution.outcome == ExpectedOutcome.WRONG_ANSWER:
        return api.SolutionTag.WA
    if solution.outcome == ExpectedOutcome.TIME_LIMIT_EXCEEDED:
        return api.SolutionTag.TL
    if solution.outcome == ExpectedOutcome.MEMORY_LIMIT_EXCEEDED:
        return api.SolutionTag.ML
    if solution.outcome == ExpectedOutcome.RUNTIME_ERROR:
        return api.SolutionTag.RE
    return api.SolutionTag.RJ


def _find_or_create_problem(problem_name: str) -> api.Problem:
    api = _get_polygon_api()
    results = api.problems_list(name=problem_name)
    for result in results:
        if result.name == problem_name:
            console.console.print(
                f'Found already existing problem [item]{problem_name}[/item].'
            )
            return result
    console.console.print(f'Creating new problem [item]{problem_name}[/item].')
    return api.problem_create(problem_name)


def _update_problem_info(problem: api.Problem):
    pkg = package.find_problem_package_or_die()

    problem.update_info(
        api.ProblemInfo(
            interactive=pkg.type == TaskType.COMMUNICATION,
            time_limit=pkg.timeLimit,
            memory_limit=pkg.memoryLimit,
        )
    )


def _get_checker_name() -> str:
    checker = package.get_checker()
    return checker.path.with_stem('checker').name


def _get_interactor_name() -> str:
    interactor = package.get_interactor()
    return interactor.path.with_stem('interactor').name


def _get_validator_name() -> str:
    validator = package.get_validator()
    return validator.path.with_stem('validator').name


def _update_rbx_header(problem: api.Problem):
    console.console.print('Uploading rbx.h...')
    rbx_header = header.get_header()
    problem.save_file(
        type=api.FileType.RESOURCE,
        name='rbx.h',
        file=rbx_header.read_bytes(),
        source_type=None,
    )


def _update_jngen(problem: api.Problem):
    jngen = download.get_jngen()
    console.console.print('Uploading jngen.h...')
    problem.save_file(
        type=api.FileType.RESOURCE,
        name='jngen.h',
        file=jngen.read_bytes(),
        source_type=None,
    )


def _update_checker(problem: api.Problem):
    checker = package.get_checker_or_builtin()
    source_type = get_polygon_language_from_code_item(checker)
    console.console.print(f'Uploading checker (lang: {source_type})...')
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_checker_name(),
        file=checker.path.read_bytes(),
        source_type=source_type,
    )

    problem.set_checker(_get_checker_name())


def _update_interactor(problem: api.Problem):
    interactor = package.get_interactor()
    source_type = get_polygon_language_from_code_item(interactor)
    console.console.print(f'Uploading interactor (lang: {source_type})...')
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_interactor_name(),
        file=interactor.path.read_bytes(),
        source_type=source_type,
    )

    problem.set_interactor(_get_interactor_name())


def _upload_validator(problem: api.Problem):
    validator = package.get_validator()
    if validator is None:
        return
    source_type = get_polygon_language_from_code_item(validator)
    console.console.print(f'Uploading validator (lang: {source_type})...')
    validator = package.get_validator()
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_validator_name(),
        file=validator.path.read_bytes(),
        source_type=source_type,
    )

    problem.set_validator(_get_validator_name())


async def _get_samples() -> List[StatementSample]:
    return await get_statement_samples(explanation_suffix='.tex')


def _save_skip_coinciding_testcases(problem: api.Problem, *args, **kwargs) -> bool:
    try:
        problem.save_test(*args, **kwargs)
    except api.PolygonRequestFailedException as e:
        if 'test coincides with' in e.comment.lower():
            return False
        raise
    return True


def _get_test_params_for_statement(
    testcase: Optional[Testcase], is_sample: bool
) -> Dict[str, Any]:
    if not is_sample or testcase is None:
        return {}
    res: Dict[str, Any] = {'test_use_in_statements': True}
    if testcase.outputPath is not None:
        res['test_output_for_statements'] = testcase.outputPath.read_text()
    else:
        return res

    interaction_path = get_best_interaction_file(testcase.outputPath)
    if interaction_path is not None:
        try:
            interaction = parse_interaction(interaction_path)
        except TestcaseInteractionParsingError:
            pass
        else:
            res['test_input_for_statements'], res['test_output_for_statements'] = (
                get_alternate_interaction_texts(interaction)
            )
            return res

    # interaction file does not exist or is not parseable, fallback to .pin and .pout.
    pin_path = testcase.outputPath.with_suffix('.pin')
    if pin_path.is_file():
        res['test_input_for_statements'] = pin_path.read_text()
    pout_path = testcase.outputPath.with_suffix('.pout')
    if pout_path.is_file():
        res['test_output_for_statements'] = pout_path.read_text()
    return res


def _get_freemarker_for_calls(calls: List[GeneratorCall], next_index: int = 1) -> str:
    return (
        '\n'.join(
            [
                f'{call.name} {call.args} > {i + next_index}'
                for i, call in enumerate(calls)
            ]
        )
        + '\n'
    )


def _upload_generator(problem: api.Problem, generator: Generator):
    generator_source_type = get_polygon_language_from_code_item(generator)
    console.console.print(
        f'Uploading generator {generator.href()} (lang: {generator_source_type})...'
    )
    try:
        problem.save_file(
            type=api.FileType.SOURCE,
            name=generator.path.name,
            file=generator.path.read_bytes(),
            source_type=generator_source_type,
        )
    except api.PolygonRequestFailedException as e:
        console.console.print(
            f'[error]Failed to upload generator {generator.href()}:[/error]\n{e.comment}'
        )
        raise typer.Exit(1) from None


def _upload_testcases(problem: api.Problem):
    entries = asyncio.run(extract_generation_testcases_from_groups())
    generators: Dict[str, Generator] = {}
    for entry in entries:
        if not entry.metadata.generator_call:
            continue
        generator = package.get_generator_or_nil(entry.metadata.generator_call.name)
        if generator is None:
            continue
        generators[str(generator.path)] = generator

    if generators:
        _update_jngen(problem)  # TODO: only upload if necessary
        console.console.print('Clearing existing script...')
        problem.save_script(testset='tests', source='<#-- empty placeholder script -->')

    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        futures = []
        for generator in generators.values():
            futures.append(executor.submit(_upload_generator, problem, generator))
        for future in futures:
            future.result()

    with rich.progress.Progress(speed_estimate_period=5) as progress:
        next_index = 1
        task_id = progress.add_task('Uploading testcases...', total=len(entries))
        calls = []
        for entry in entries:
            if entry.metadata.generator_call is not None:
                # Generated testcases are handled later.
                generator = package.get_generator_or_nil(
                    entry.metadata.generator_call.name
                )
                if generator is None:
                    continue
                calls.append(
                    GeneratorCall(
                        name=generator.path.stem,
                        args=entry.metadata.generator_call.args,
                    )
                )
                continue

            content = entry.metadata.content
            if (
                entry.metadata.copied_from is not None
                and entry.metadata.copied_from.inputPath.is_file()
            ):
                content = entry.metadata.copied_from.inputPath.read_text()
            if content is None:
                continue
            saved = _save_skip_coinciding_testcases(
                problem,
                testset='tests',
                test_index=next_index,
                test_input=content,
                **_get_test_params_for_statement(
                    entry.metadata.copied_from,
                    is_sample=entry.is_sample(),
                ),
            )
            progress.update(task_id, advance=1)
            if saved:
                next_index += 1

        if calls:
            try:
                problem.save_script(
                    testset='tests', source=_get_freemarker_for_calls(calls, next_index)
                )
            except api.PolygonRequestFailedException as e:
                if 'already used in non-script' in e.comment.lower():
                    console.console.print(f'[error]{e.comment}[/error]')
                    console.console.print(
                        '[error]Please remove the conflicting manual tests on the Polygon UI and try again.[/error]'
                    )
                    raise typer.Exit(1) from None
                raise
        progress.update(task_id, completed=len(entries))


def _upload_solutions(problem: api.Problem):
    saved_solutions = set()

    def process_solution(solution: Solution, i: int):
        source_type = get_polygon_language_from_code_item(solution)
        console.console.print(
            f'Uploading solution {solution.href()} (lang: {source_type}, tag: [item]{_get_solution_tag(solution, is_first=i == 0)}[/item])...'
        )
        problem.save_solution(
            solution.path.name,
            solution.path.read_bytes(),
            source_type=source_type,
            tag=_get_solution_tag(solution, is_first=i == 0),
        )
        saved_solutions.add(solution.path.name)

    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        futures = []
        for i, solution in enumerate(package.get_solutions()):
            futures.append(executor.submit(process_solution, solution, i))
        for solution, future in zip(package.get_solutions(), futures):
            try:
                future.result()
            except api.PolygonRequestFailedException as e:
                console.console.print(
                    f'[error]Failed to upload solution {solution.href()}:[/error]\n{e.comment}'
                )

    def delete_solution(solution: api.Solution):
        console.console.print(f'Deleting solution [item]{solution.name}[/item]...')
        problem.save_solution(
            solution.name,
            '# This solution is no longer used in the problem but was kept by rbx.cp.',
            source_type='python.3',
            tag=api.SolutionTag.NR,
        )

    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        futures = []
        solutions = [
            solution
            for solution in problem.solutions()
            if solution.name not in saved_solutions
        ]
        for solution in solutions:
            futures.append(executor.submit(delete_solution, solution))
        for solution, future in zip(solutions, futures):
            try:
                future.result()
            except api.PolygonRequestFailedException as e:
                console.console.print(
                    f'[error]Failed to delete solution [item]{solution.name}[/item]:[/error]\n{e.comment}'  # pyright: ignore[reportAttributeAccessIssue]
                )


def _get_explanations(explanations: Dict[int, str]) -> str:
    entries = []
    for i in sorted(explanations):
        explanation = explanations[i]
        entries.append(f'\\textbf{{Explanation for example {i + 1}}}\n\n{explanation}')
    return '\n\n'.join(entries)


def _upload_statement_resources(
    problem: api.Problem, statement: Statement
) -> Dict[str, str]:
    res: Dict[str, str] = {}
    assets = get_relative_assets(statement.path, statement.assets)
    assets.extend(get_produced_tikz_pdfs(statement))
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        futures = []
        for asset, relative_asset in assets:
            console.console.print(
                f'Uploading statement resource [item]{relative_asset}[/item]...'
            )
            resource_bytes = asset.read_bytes()
            if len(resource_bytes) >= 1024 * 1024:  # >= 1mb
                console.console.print(
                    f'[error]Statement resource [item]{relative_asset}[/item] is too large to upload (more than 1MB).[/error]'
                )
                raise typer.Exit(1)
            no_suffix_relative_asset = relative_asset.with_suffix('')
            key_asset = (
                None
                if len(no_suffix_relative_asset.parents) <= 1
                else str(no_suffix_relative_asset).replace('/', '__')
                + relative_asset.suffix
            )
            if key_asset is not None:
                res[str(relative_asset.with_suffix(''))] = key_asset
            if key_asset is None:
                key_asset = str(no_suffix_relative_asset) + relative_asset.suffix
            console.console.print(
                f'Uploading statement resource [item]{relative_asset}[/item] (normalized name: [item]{key_asset}[/item])...'
            )
            futures.append(
                executor.submit(
                    problem.save_statement_resource,
                    name=key_asset,
                    file=resource_bytes,
                )
            )
    for future in futures:
        future.result()
    return res


async def _upload_statement(
    problem: api.Problem, main_language: Optional[str], upload_as_english: bool = False
):
    pkg = package.find_problem_package_or_die()

    def process_statement(statement: Statement, language: str, uploaded_language: str):
        console.console.print(
            f'Uploading statement for language [item]{language}[/item] (uploaded language: [item]{uploaded_language}[/item])...'
        )
        blocks = get_processed_statement_blocks(statement)
        resources = _upload_statement_resources(problem, statement)

        def _replace_resources(block: str, resources=resources) -> str:
            for key, value in resources.items():
                block = block.replace(key, value)
            return block

        def _get_block(block_name: str, blocks=blocks, resources=resources) -> str:
            block = blocks.blocks.get(block_name) or ''
            return _replace_resources(block, resources)

        def _get_notes_with_explanations(
            blocks=blocks,
        ) -> Optional[str]:
            notes = _get_block('notes')
            if notes is None and not blocks.explanations:
                return None
            res = _replace_resources(_get_explanations(blocks.explanations))
            if notes is not None:
                res = notes + '\n\n' + res
            return res

        polygon_statement = api.Statement(
            encoding='utf-8',
            name=naming.get_problem_title(statement.language, statement, pkg),
            legend=_get_block('legend'),
            input=_get_block('input'),
            output=_get_block('output'),
            interaction=_get_block('interaction')
            if pkg.type == TaskType.COMMUNICATION
            else None,
            notes=_get_notes_with_explanations() or '',
        )
        problem.save_statement(
            lang=uploaded_language,
            problem_statement=polygon_statement,
        )

    process_statements(main_language, upload_as_english, process_statement)


def _normalize_problem_name(name: str) -> str:
    return name.replace(' ', '-').replace('_', '-').lower()


async def upload_problem(
    name: str,
    main_language: Optional[str],
    upload_as_english: bool = False,
    upload_only: Optional[Set[str]] = None,
    dont_upload: Optional[Set[str]] = None,
):
    if upload_only is None:
        upload_only = set()
    if dont_upload is None:
        dont_upload = set()

    if not upload_only:
        upload_only = set(ALL_PARAMS_CHOICES)

    which_upload = upload_only - dont_upload
    pkg = package.find_problem_package_or_die()
    name = _normalize_problem_name(name)
    problem = _find_or_create_problem(name)
    _update_problem_info(problem)

    if 'files' in which_upload:
        _update_rbx_header(problem)
        _update_checker(problem)

    if (
        pkg.type == TaskType.COMMUNICATION
        and package.get_interactor_or_nil() is not None
    ):
        if 'files' in which_upload:
            _update_interactor(problem)

    if pkg.validator is not None:
        if 'files' in which_upload:
            _upload_validator(problem)

    if 'solutions' in which_upload:
        _upload_solutions(problem)
    if 'tests' in which_upload:
        _upload_testcases(problem)
    if 'statements' in which_upload:
        await _upload_statement(
            problem, main_language=main_language, upload_as_english=upload_as_english
        )

    # Commit.
    console.console.print('Committing changes...')
    problem.commit_changes()

    console.console.print(
        f'[success]Problem [item]{name}[/item] uploaded successfully![/success]'
    )



import contextvars
import enum
from collections import OrderedDict
from typing import Callable, List, Optional, Tuple, Union

from rbx import console


class IssueLevel(enum.Enum):
    OVERVIEW = enum.auto()
    DETAILED = enum.auto()


class IssueSeverity(enum.Enum):
    ERROR = enum.auto()
    WARNING = enum.auto()


class Issue:
    def get_detailed_section(self) -> Optional[Tuple[str, ...]]:
        return None

    def get_overview_section(self) -> Optional[Tuple[str, ...]]:
        return None

    def get_detailed_message(self) -> str:
        return ''

    def get_overview_message(self) -> str:
        return ''

    def get_severity(self) -> IssueSeverity:
        return IssueSeverity.ERROR


IssueSection = OrderedDict[str, Union[List[Issue], 'IssueSection']]


class IssueAccumulator:
    def __init__(self):
        self.issues = []

    def add_issue(self, issue: Issue):
        self.issues.append(issue)

    def get_sections_by(
        self, key: Callable[[Issue], Optional[Tuple[str, ...]]]
    ) -> IssueSection:
        sections = OrderedDict()
        for issue in self.issues:
            section_key = key(issue)
            if section_key is None:
                continue
            current = sections
            for k in section_key[:-1]:
                current = current.setdefault(k, OrderedDict())
            current.setdefault(section_key[-1], [])
            current[section_key[-1]].append(issue)

        return sections

    def get_detailed_sections(self) -> IssueSection:
        return self.get_sections_by(lambda issue: issue.get_detailed_section())

    def get_overview_sections(self) -> IssueSection:
        return self.get_sections_by(lambda issue: issue.get_overview_section())

    def _print_report_by(
        self,
        section_fn: Callable[[], IssueSection],
        message_fn: Callable[[Issue], str],
    ):
        from rich.tree import Tree

        tree = Tree('Issues')
        sections = section_fn()

        def print_section(section: IssueSection, tree: Tree):
            for key, value in section.items():
                child = tree.add(key)
                if isinstance(value, OrderedDict):
                    print_section(value, child)
                    continue
                seen_issues = set()
                for issue in value:
                    severity = issue.get_severity()
                    msg = message_fn(issue)
                    if msg in seen_issues:
                        continue
                    seen_issues.add(msg)
                    if severity == IssueSeverity.ERROR:
                        child.add(f'[error]{msg}[/error]')
                    elif severity == IssueSeverity.WARNING:
                        child.add(f'[warning]{msg}[/warning]')

        print_section(sections, tree)

        if tree.children:
            console.console.rule('[error]Issues[/error]', style='error')
            for child in tree.children:
                console.console.print(child)

    def print_detailed_report(self):
        self._print_report_by(
            self.get_detailed_sections, lambda issue: issue.get_detailed_message()
        )

    def print_overview_report(self):
        self._print_report_by(
            self.get_overview_sections, lambda issue: issue.get_overview_message()
        )


issue_stack_var: contextvars.ContextVar[Optional[List[IssueAccumulator]]] = (
    contextvars.ContextVar('issue_stack', default=[IssueAccumulator()])
)
issue_level_var = contextvars.ContextVar('issue_level', default=IssueLevel.DETAILED)


def get_issue_stack() -> list[IssueAccumulator]:
    stack = issue_stack_var.get()
    if stack is None:
        stack = [IssueAccumulator()]
        issue_stack_var.set(stack)
    return stack


def push_issue_accumulator():
    issue_stack_var.set(get_issue_stack() + [IssueAccumulator()])


def pop_issue_accumulator():
    issue_stack_var.set(get_issue_stack()[:-1])


def get_issue_accumulator() -> IssueAccumulator:
    return get_issue_stack()[-1]


def add_issue(issue: Issue):
    for acc in get_issue_stack():
        acc.add_issue(issue)


def print_current_report():
    acc = get_issue_accumulator()
    if issue_level_var.get() == IssueLevel.OVERVIEW:
        acc.print_overview_report()
    else:
        acc.print_detailed_report()



import pytest
import typer

from rbx.box.generators import generate_outputs_for_testcases, generate_testcases
from rbx.box.schema import ExpectedOutcome
from rbx.box.testcase_utils import TestcaseEntry
from rbx.box.testing import testing_package
from rbx.grading import steps

_SOL_DOUBLE_NUMBER = """
#include <iostream>

using namespace std;

int main() {
  int x; cin >> x;
  cout << x * 2 << endl;
}
"""


_SOL_COMPILE_ERROR = """
#include <iostream>

// Missing std;

int main() {
  int x; cin >> x;
  cout << x * 2 << endl;
}
"""


_SOL_TRIPLE_NUMBER = """
#include <iostream>

using namespace std;

int main() {
  int x; cin >> x;
  cout << x * 3 << endl;
}
"""


_SOL_QUADRUPLE_NUMBER = """
#include <iostream>

using namespace std;

int main() {
  int x; cin >> x;
  cout << x * 4 << endl;
}
"""


async def test_generator_outputs_are_generated(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen.cpp 123')
    testing_pkg.add_solution('sol.cpp', outcome=ExpectedOutcome.ACCEPTED).write_text(
        _SOL_DOUBLE_NUMBER
    )

    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated.
    entries = [TestcaseEntry(group='main', index=0)]
    await generate_outputs_for_testcases(entries)

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.out'
    ).read_bytes() == b'246\n'


async def test_generator_outputs_no_main_solution_and_needs_output(
    testing_pkg: testing_package.TestingPackage,
    capsys: pytest.CaptureFixture[str],
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen.cpp 123')
    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated.
    entries = [TestcaseEntry(group='main', index=0)]

    with pytest.raises(typer.Exit):
        await generate_outputs_for_testcases(entries)

    out = capsys.readouterr().out
    assert 'No main/model solution found' in out


async def test_generator_outputs_no_main_solution_and_does_not_need_output(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_glob('main', 'manual_tests/*.in')
    testing_pkg.add_file('manual_tests/000.in').write_text('123\n')
    testing_pkg.add_file('manual_tests/000.ans').write_text('246\n')
    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated.
    entries = [TestcaseEntry(group='main', index=0)]

    await generate_outputs_for_testcases(entries)

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.out'
    ).read_bytes() == b'246\n'


async def test_generator_outputs_main_solution_does_not_compile_and_needs_output(
    testing_pkg: testing_package.TestingPackage,
    capsys: pytest.CaptureFixture[str],
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen.cpp 123')
    testing_pkg.add_solution('sol.cpp', outcome=ExpectedOutcome.ACCEPTED).write_text(
        _SOL_COMPILE_ERROR
    )

    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated.
    entries = [TestcaseEntry(group='main', index=0)]

    with pytest.raises(steps.CompilationError):
        await generate_outputs_for_testcases(entries)

    out = capsys.readouterr().out
    assert 'Failed compiling main solution' in out


async def test_generator_outputs_main_solution_does_not_compile_and_does_not_need_output(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_glob('main', 'manual_tests/*.in')
    testing_pkg.add_file('manual_tests/000.in').write_text('123\n')
    testing_pkg.add_file('manual_tests/000.ans').write_text('246\n')
    testing_pkg.add_solution('sol.cpp', outcome=ExpectedOutcome.ACCEPTED).write_text(
        _SOL_COMPILE_ERROR
    )
    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated.
    entries = [TestcaseEntry(group='main', index=0)]

    await generate_outputs_for_testcases(entries)

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.out'
    ).read_bytes() == b'246\n'


async def test_generator_outputs_with_model_solution_in_samples_group(
    testing_pkg: testing_package.TestingPackage,
):
    """Test that model solution is used when specified in samples testgroup."""
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('samples', 'gens/gen.cpp 123')

    # Add main solution that doubles
    testing_pkg.add_solution('sol.cpp', outcome=ExpectedOutcome.ACCEPTED).write_text(
        _SOL_DOUBLE_NUMBER
    )

    # Add model solution that triples - should be used for samples group
    testing_pkg.add_file('model_sol.cpp').write_text(_SOL_TRIPLE_NUMBER)

    # Manually configure samples group with model solution
    import pathlib

    from rbx.box.schema import Solution

    samples_group = testing_pkg.yml.testcases[0]
    samples_group.model_solution = Solution(path=pathlib.Path('model_sol.cpp'))
    testing_pkg.save()

    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated.
    entries = [TestcaseEntry(group='samples', index=0)]
    await generate_outputs_for_testcases(entries)

    # Should use model solution (triple) instead of main solution (double)
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '000.out'
    ).read_bytes() == b'369\n'


async def test_generator_outputs_with_multiple_model_solutions(
    testing_pkg: testing_package.TestingPackage,
):
    """Test compilation of multiple different model solutions."""
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')

    # Add main solution that doubles
    testing_pkg.add_solution('sol.cpp', outcome=ExpectedOutcome.ACCEPTED).write_text(
        _SOL_DOUBLE_NUMBER
    )

    # Add different model solutions
    testing_pkg.add_file('model_sol1.cpp').write_text(_SOL_TRIPLE_NUMBER)
    testing_pkg.add_file('model_sol2.cpp').write_text(_SOL_QUADRUPLE_NUMBER)

    # Configure multiple testgroups with different model solutions
    import pathlib

    from rbx.box.schema import Solution

    testing_pkg.add_testgroup_from_plan('samples', 'gens/gen.cpp 123')
    samples_group = testing_pkg.yml.testcases[0]
    samples_group.model_solution = Solution(path=pathlib.Path('model_sol1.cpp'))

    testing_pkg.add_testgroup_from_plan('group2', 'gens/gen.cpp 456')
    # Note: group2 will use main solution since only samples can have model_solution

    testing_pkg.save()

    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated.
    entries = [
        TestcaseEntry(group='samples', index=0),
        TestcaseEntry(group='group2', index=0),
    ]
    await generate_outputs_for_testcases(entries)

    # samples should use model solution (triple)
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '000.out'
    ).read_bytes() == b'369\n'

    # group2 should use main solution (double)
    assert (
        testing_pkg.get_build_testgroup_path('group2') / '000.out'
    ).read_bytes() == b'912\n'


async def test_generator_outputs_model_solution_compilation_failure(
    testing_pkg: testing_package.TestingPackage,
    capsys: pytest.CaptureFixture[str],
):
    """Test handling of model solution compilation failure."""
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('samples', 'gens/gen.cpp 123')

    # Add main solution that compiles
    testing_pkg.add_solution('sol.cpp', outcome=ExpectedOutcome.ACCEPTED).write_text(
        _SOL_DOUBLE_NUMBER
    )

    # Add model solution that doesn't compile
    testing_pkg.add_file('model_sol.cpp').write_text(_SOL_COMPILE_ERROR)

    # Configure samples group with broken model solution
    import pathlib

    from rbx.box.schema import Solution

    samples_group = testing_pkg.yml.testcases[0]
    samples_group.model_solution = Solution(path=pathlib.Path('model_sol.cpp'))
    testing_pkg.save()

    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs generation fails with proper error
    entries = [TestcaseEntry(group='samples', index=0)]

    with pytest.raises(steps.CompilationError):
        await generate_outputs_for_testcases(entries)

    out = capsys.readouterr().out
    assert 'Failed compiling model solution' in out
    assert 'model_sol.cpp' in out


async def test_generator_outputs_model_solution_reused_across_testcases(
    testing_pkg: testing_package.TestingPackage,
):
    """Test that model solution is compiled once and reused across multiple testcases."""
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan(
        'samples',
        """
gens/gen.cpp 123
gens/gen.cpp 456
gens/gen.cpp 789
""",
    )

    # Add main solution that doubles
    testing_pkg.add_solution('sol.cpp', outcome=ExpectedOutcome.ACCEPTED).write_text(
        _SOL_DOUBLE_NUMBER
    )

    # Add model solution that triples
    testing_pkg.add_file('model_sol.cpp').write_text(_SOL_TRIPLE_NUMBER)

    # Configure samples group with model solution
    import pathlib

    from rbx.box.schema import Solution

    samples_group = testing_pkg.yml.testcases[0]
    samples_group.model_solution = Solution(path=pathlib.Path('model_sol.cpp'))
    testing_pkg.save()

    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated for all testcases
    entries = [
        TestcaseEntry(group='samples', index=0),
        TestcaseEntry(group='samples', index=1),
        TestcaseEntry(group='samples', index=2),
    ]
    await generate_outputs_for_testcases(entries)

    # All should use model solution (triple)
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '000.out'
    ).read_bytes() == b'369\n'
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '001.out'
    ).read_bytes() == b'1368\n'
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '002.out'
    ).read_bytes() == b'2367\n'


async def test_generator_outputs_fallback_to_main_when_no_model_solution(
    testing_pkg: testing_package.TestingPackage,
):
    """Test that main solution is used when no model solution is specified."""
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('samples', 'gens/gen.cpp 123')

    # Add main solution that doubles
    testing_pkg.add_solution('sol.cpp', outcome=ExpectedOutcome.ACCEPTED).write_text(
        _SOL_DOUBLE_NUMBER
    )

    # Don't add model solution - should fallback to main solution

    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs are generated.
    entries = [TestcaseEntry(group='samples', index=0)]
    await generate_outputs_for_testcases(entries)

    # Should use main solution (double)
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '000.out'
    ).read_bytes() == b'246\n'


async def test_generator_outputs_no_main_or_model_solution_error_message(
    testing_pkg: testing_package.TestingPackage,
    capsys: pytest.CaptureFixture[str],
):
    """Test updated error message when neither main nor model solution is available."""
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen.cpp 123')

    # Don't add any solution

    # Ensure cases are generated.
    await generate_testcases()

    # Ensure outputs generation fails with updated error message
    entries = [TestcaseEntry(group='main', index=0)]

    with pytest.raises(typer.Exit):
        await generate_outputs_for_testcases(entries)

    out = capsys.readouterr().out
    # Check that the error message mentions both main and model solutions
    assert 'No main/model solution found' in out



import os
import signal
import subprocess
import sys
import time
from unittest import mock

import pytest

from rbx.grading.judge.program import (
    Program,
    ProgramCode,
    ProgramError,
    ProgramIO,
    ProgramParams,
    ProgramResult,
    get_cpu_time,
    get_memory_usage,
    get_preexec_fn,
)


@pytest.fixture
def program_test_data(testdata_path):
    """Path to program test data directory."""
    return testdata_path / 'program_test'


@pytest.fixture
def simple_hello_program(program_test_data):
    """Path to simple hello world program."""
    return program_test_data / 'simple_hello.py'


@pytest.fixture
def infinite_loop_program(program_test_data):
    """Path to infinite loop program."""
    return program_test_data / 'infinite_loop.py'


@pytest.fixture
def memory_hog_program(program_test_data):
    """Path to memory-intensive program."""
    return program_test_data / 'memory_hog.py'


@pytest.fixture
def output_large_program(program_test_data):
    """Path to large output program."""
    return program_test_data / 'output_large.py'


@pytest.fixture
def exit_with_code_program(program_test_data):
    """Path to program that exits with specific codes."""
    return program_test_data / 'exit_with_code.py'


@pytest.fixture
def signal_test_program(program_test_data):
    """Path to signal handling program."""
    return program_test_data / 'signal_test.py'


@pytest.fixture
def input_echo_program(program_test_data):
    """Path to input echo program."""
    return program_test_data / 'input_echo.py'


@pytest.fixture
def cpu_intensive_program(program_test_data):
    """Path to CPU-intensive program."""
    return program_test_data / 'cpu_intensive.py'


class TestProgramIO:
    """Test ProgramIO class functionality."""

    def test_default_io(self):
        """Test default ProgramIO settings."""
        io = ProgramIO()
        assert io.input == subprocess.PIPE
        assert io.output == subprocess.PIPE
        assert io.stderr == subprocess.PIPE

    def test_custom_io_paths(self, tmp_path):
        """Test ProgramIO with file paths."""
        input_file = tmp_path / 'input.txt'
        output_file = tmp_path / 'output.txt'
        stderr_file = tmp_path / 'stderr.txt'

        input_file.write_text('test input\n')

        io = ProgramIO(
            input=str(input_file), output=str(output_file), stderr=str(stderr_file)
        )

        input_fobj, output_fobj, stderr_fobj = io.get_file_objects()

        # Test that files are opened correctly
        assert hasattr(input_fobj, 'read')
        assert hasattr(output_fobj, 'write')
        assert hasattr(stderr_fobj, 'write')

        # Close the files only if they're not file descriptors
        if not isinstance(input_fobj, int):
            input_fobj.close()
        if not isinstance(output_fobj, int):
            output_fobj.close()
        if not isinstance(stderr_fobj, int):
            stderr_fobj.close()

    def test_io_with_file_descriptors(self):
        """Test ProgramIO with file descriptors."""
        io = ProgramIO(input=0, output=1, stderr=2)
        input_fobj, output_fobj, stderr_fobj = io.get_file_objects()

        assert input_fobj == 0
        assert output_fobj == 1
        assert stderr_fobj == 2

    def test_io_creates_parent_directories(self, tmp_path):
        """Test that ProgramIO creates parent directories for output files."""
        nested_output = tmp_path / 'nested' / 'dir' / 'output.txt'
        nested_stderr = tmp_path / 'nested' / 'dir' / 'stderr.txt'

        io = ProgramIO(output=str(nested_output), stderr=str(nested_stderr))
        input_fobj, output_fobj, stderr_fobj = io.get_file_objects()

        assert nested_output.parent.exists()
        assert nested_stderr.parent.exists()

        # Close the files only if they're not file descriptors
        if not isinstance(output_fobj, int):
            output_fobj.close()
        if not isinstance(stderr_fobj, int):
            stderr_fobj.close()


class TestProgramParams:
    """Test ProgramParams class functionality."""

    def test_default_params(self):
        """Test default ProgramParams settings."""
        params = ProgramParams()
        assert isinstance(params.io, ProgramIO)
        assert params.chdir is None
        assert params.time_limit is None
        assert params.wall_time_limit is None
        assert params.memory_limit is None
        assert params.fs_limit is None
        assert params.env == {}
        assert params.pgid is None

    def test_custom_params(self, tmp_path):
        """Test custom ProgramParams settings."""
        params = ProgramParams(
            chdir=tmp_path,
            time_limit=5.0,
            wall_time_limit=10.0,
            memory_limit=100,
            fs_limit=1000,
            env={'TEST_VAR': 'test_value'},
            pgid=12345,
        )

        assert params.chdir == tmp_path
        assert params.time_limit == 5.0
        assert params.wall_time_limit == 10.0
        assert params.memory_limit == 100
        assert params.fs_limit == 1000
        assert params.env == {'TEST_VAR': 'test_value'}
        assert params.pgid == 12345


class TestPreexecFn:
    """Test preexec_fn functionality."""

    def test_preexec_fn_with_limits(self):
        """Test that preexec_fn is created correctly with limits."""
        params = ProgramParams(time_limit=5.0, fs_limit=1000, pgid=12345)

        preexec_fn = get_preexec_fn(params)
        assert callable(preexec_fn)

    def test_preexec_fn_no_limits(self):
        """Test that preexec_fn works without limits."""
        params = ProgramParams()
        preexec_fn = get_preexec_fn(params)
        assert callable(preexec_fn)


class TestResourceUtilities:
    """Test resource utility functions."""

    def test_get_cpu_time(self):
        """Test get_cpu_time function."""
        # Create a mock resource usage object
        mock_ru = mock.MagicMock()
        mock_ru.ru_utime = 1.5
        mock_ru.ru_stime = 0.5

        cpu_time = get_cpu_time(mock_ru)
        assert cpu_time == 2.0

    @pytest.mark.parametrize(
        'platform,expected_multiplier',
        [
            ('darwin', 1),
            ('linux', 1024),
        ],
    )
    def test_get_memory_usage(self, platform, expected_multiplier):
        """Test get_memory_usage function for different platforms."""
        mock_ru = mock.MagicMock()
        mock_ru.ru_maxrss = 1000
        mock_ru.ru_ixrss = 100
        mock_ru.ru_idrss = 50
        mock_ru.ru_isrss = 25

        with mock.patch('sys.platform', platform):
            memory_usage = get_memory_usage(mock_ru)

            if platform == 'darwin':
                # On macOS: ru_maxrss is in bytes, ru_ixrss in KB
                expected = 1000 + 100 * 1024
            else:
                # On Linux: all values in KB, convert to bytes
                expected = (1000 + 100 + 50 + 25) * 1024

            assert memory_usage == expected


class TestProgram:
    """Test Program class functionality."""

    def test_simple_program_execution(self, simple_hello_program):
        """Test basic program execution."""
        command = [sys.executable, str(simple_hello_program)]
        params = ProgramParams()

        program = Program(command, params)
        result = program.wait()

        assert isinstance(result, ProgramResult)
        assert result.exitcode == 0
        assert result.wall_time > 0
        assert result.cpu_time >= 0
        assert result.memory_used > 0
        assert len(result.program_codes) == 0  # No error codes for successful execution

    def test_program_with_exit_code(self, exit_with_code_program):
        """Test program execution with non-zero exit code."""
        command = [sys.executable, str(exit_with_code_program), '42']
        params = ProgramParams()

        program = Program(command, params)
        result = program.wait()

        assert result.exitcode == 42
        assert ProgramCode.RE in result.program_codes

    def test_program_with_input_output(self, input_echo_program, tmp_path):
        """Test program execution with input and output files."""
        input_file = tmp_path / 'input.txt'
        output_file = tmp_path / 'output.txt'
        stderr_file = tmp_path / 'stderr.txt'

        input_file.write_text('Hello\nWorld\n')

        io = ProgramIO(
            input=str(input_file), output=str(output_file), stderr=str(stderr_file)
        )
        params = ProgramParams(io=io)
        command = [sys.executable, str(input_echo_program)]

        program = Program(command, params)
        result = program.wait()

        assert result.exitcode == 0
        assert output_file.exists()
        assert stderr_file.exists()

        output_content = output_file.read_text()
        assert 'Echo: Hello' in output_content
        assert 'Echo: World' in output_content

    def test_program_with_working_directory(self, simple_hello_program, tmp_path):
        """Test program execution with custom working directory."""
        params = ProgramParams(chdir=tmp_path)
        command = [sys.executable, str(simple_hello_program)]

        program = Program(command, params)
        result = program.wait()

        assert result.exitcode == 0

    def test_program_with_environment_variables(self, tmp_path):
        """Test program execution with custom environment variables."""
        test_script = tmp_path / 'env_test.py'
        output_file = tmp_path / 'output.txt'

        test_script.write_text(
            "import os; print(os.environ.get('TEST_VAR', 'NOT_FOUND'))"
        )

        io = ProgramIO(output=str(output_file))
        params = ProgramParams(io=io, env={'TEST_VAR': 'test_value'})
        command = [sys.executable, str(test_script)]

        program = Program(command, params)
        result = program.wait()

        assert result.exitcode == 0

        # Verify the environment variable was correctly passed
        output_content = output_file.read_text().strip()
        assert output_content == 'test_value'

    def test_cpu_time_limit(self, cpu_intensive_program):
        """Test CPU time limit enforcement."""
        params = ProgramParams(time_limit=0.1)  # Very short time limit
        command = [sys.executable, str(cpu_intensive_program)]

        program = Program(command, params)
        result = program.wait()

        # Should be killed due to CPU time limit
        assert ProgramCode.TO in result.program_codes

    def test_wall_time_limit(self, infinite_loop_program):
        """Test wall time limit enforcement."""
        params = ProgramParams(wall_time_limit=0.5)  # Short wall time limit
        command = [sys.executable, str(infinite_loop_program)]

        program = Program(command, params)
        result = program.wait()

        # Should be killed due to wall time limit
        assert ProgramCode.WT in result.program_codes
        assert ProgramCode.TO in result.program_codes
        assert result.wall_time >= 0.5

    def test_memory_limit(self, memory_hog_program):
        """Test memory limit enforcement."""
        params = ProgramParams(memory_limit=10)  # 10 MB limit
        command = [sys.executable, str(memory_hog_program)]

        program = Program(command, params)
        result = program.wait()

        # Should be killed due to memory limit
        assert ProgramCode.ML in result.program_codes

    def test_output_limit(self, output_large_program, tmp_path):
        """Test file size limit enforcement."""
        output_file = tmp_path / 'output.txt'
        io = ProgramIO(output=str(output_file))
        params = ProgramParams(io=io, fs_limit=1)  # 1 KB limit
        command = [sys.executable, str(output_large_program)]

        program = Program(command, params)
        result = program.wait()

        # Should exceed file size limit
        assert ProgramCode.OL in result.program_codes

    def test_signal_handling(self, signal_test_program):
        """Test signal handling and program termination."""
        params = ProgramParams()
        command = [sys.executable, str(signal_test_program)]

        program = Program(command, params)

        # Give the program a moment to start
        time.sleep(1.0)

        # Send SIGTERM to the process
        os.kill(program.pid, signal.SIGTERM)

        result = program.wait()

        # Two possible outcomes due to race condition:

        # 1. Signal handler executes: exit code 143 = 128 + SIGTERM (15)
        # 2. Process killed by signal: exit code -15
        assert result.exitcode in [143, -15]

        # Both cases should be marked as errors
        if result.exitcode == 143:
            assert ProgramCode.RE in result.program_codes
        else:  # result.exitcode == -15
            assert ProgramCode.SG in result.program_codes
            assert result.killing_signal == 15

    def test_program_pipes_access(self, simple_hello_program):
        """Test access to program pipes."""
        params = ProgramParams()
        command = [sys.executable, str(simple_hello_program)]

        program = Program(command, params)
        pipes = program.pipes

        assert pipes.input is not None
        assert pipes.output is not None
        assert pipes.stderr is not None

        result = program.wait()
        assert result.exitcode == 0

    def test_program_pid_access(self, simple_hello_program):
        """Test access to program PID."""
        params = ProgramParams()
        command = [sys.executable, str(simple_hello_program)]

        program = Program(command, params)
        pid = program.pid

        assert isinstance(pid, int)
        assert pid > 0

        result = program.wait()
        assert result.exitcode == 0

    def test_multiple_program_codes(self, tmp_path):
        """Test that multiple program codes can be set simultaneously."""
        # Create a program that will exceed both time and memory limits
        test_script = tmp_path / 'multi_limit_test.py'
        test_script.write_text("""
import time
data = []
for i in range(1000000):
    data.append([0] * 1000)
    time.sleep(0.001)
""")

        params = ProgramParams(
            time_limit=0.1,
            memory_limit=1,  # 1 MB
            wall_time_limit=0.2,
        )
        command = [sys.executable, str(test_script)]

        program = Program(command, params)
        result = program.wait()

        # Could have multiple violation codes
        violation_codes = [
            code
            for code in result.program_codes
            if code in [ProgramCode.TO, ProgramCode.WT, ProgramCode.ML]
        ]
        assert len(violation_codes) > 0

    def test_program_result_attributes(self, simple_hello_program):
        """Test that ProgramResult has all expected attributes."""
        params = ProgramParams()
        command = [sys.executable, str(simple_hello_program)]

        program = Program(command, params)
        result = program.wait()

        # Check all attributes exist and have correct types
        assert isinstance(result.exitcode, int)
        assert isinstance(result.wall_time, float)
        assert isinstance(result.cpu_time, float)
        assert isinstance(result.memory_used, int)
        assert isinstance(result.file_sizes, int)
        assert isinstance(result.program_codes, list)
        assert result.killing_signal is None or isinstance(result.killing_signal, int)
        assert result.alarm_msg is None or isinstance(result.alarm_msg, str)

    def test_file_size_calculation(self, tmp_path):
        """Test file size calculation for output files."""
        output_file = tmp_path / 'output.txt'
        stderr_file = tmp_path / 'stderr.txt'

        test_script = tmp_path / 'file_size_test.py'
        test_script.write_text("""
print("Hello" * 100)
import sys
print("Error" * 50, file=sys.stderr)
""")

        io = ProgramIO(output=str(output_file), stderr=str(stderr_file))
        params = ProgramParams(io=io)
        command = [sys.executable, str(test_script)]

        program = Program(command, params)
        result = program.wait()

        assert result.exitcode == 0
        assert result.file_sizes > 0

        # Verify actual file sizes match
        actual_size = output_file.stat().st_size + stderr_file.stat().st_size
        assert result.file_sizes == actual_size


class TestProgramCode:
    """Test ProgramCode enum."""

    def test_program_code_values(self):
        """Test that ProgramCode enum has expected values."""
        assert ProgramCode.RE.value == 'RE'  # Runtime Error
        assert ProgramCode.SG.value == 'SG'  # Signal
        assert ProgramCode.TO.value == 'TO'  # Time Out
        assert ProgramCode.WT.value == 'WT'  # Wall Time
        assert ProgramCode.ML.value == 'ML'  # Memory Limit
        assert ProgramCode.OL.value == 'OL'  # Output Limit
        assert ProgramCode.TE.value == 'TE'  # Time Error


class TestProgramMocks:
    """Test Program functionality using mocks to verify intended behavior."""

    @mock.patch('rbx.grading.judge.program.get_file_sizes')
    @mock.patch('rbx.grading.judge.program.get_memory_usage')
    @mock.patch('rbx.grading.judge.program.get_cpu_time')
    @mock.patch('os.wait4')
    def test_process_exit_with_mocks(
        self,
        mock_wait4,
        mock_get_cpu_time,
        mock_get_memory_usage,
        mock_get_file_sizes,
        simple_hello_program,
    ):
        """Test that process_exit correctly processes resource usage data."""
        # Setup mocks
        mock_wait4.return_value = (12345, 0, mock.MagicMock())  # pid, exitstatus, ru
        mock_get_cpu_time.return_value = 1.5
        mock_get_memory_usage.return_value = 1024 * 1024  # 1MB in bytes
        mock_get_file_sizes.return_value = 2048

        command = [sys.executable, str(simple_hello_program)]
        params = ProgramParams()
        program = Program(command, params)

        # Wait for the process to complete and get the mocked result
        result = program.wait()

        # Verify that our utility functions were called
        mock_get_cpu_time.assert_called_once()
        mock_get_memory_usage.assert_called_once()
        mock_get_file_sizes.assert_called_once_with(params.io)

        # Verify the result uses mocked values
        assert result.cpu_time == 1.5
        assert result.memory_used == 1024 * 1024
        assert result.file_sizes == 2048

    @mock.patch('os.setpgid')
    @mock.patch('resource.setrlimit')
    @mock.patch('sys.platform', 'linux')
    def test_preexec_fn_behavior_linux(self, mock_setrlimit, mock_setpgid):
        """Test that preexec_fn actually calls the expected system functions."""
        import resource as resource_module

        params = ProgramParams(time_limit=5.0, fs_limit=1000, pgid=12345)

        preexec_fn = get_preexec_fn(params)
        preexec_fn()

        # Verify setpgid was called
        mock_setpgid.assert_called_once_with(0, 12345)

        # Verify setrlimit was called 3 times.
        assert mock_setrlimit.call_count == 3
        calls = mock_setrlimit.call_args_list

        # Check CPU limit call
        cpu_call = calls[0]
        assert cpu_call[0][0] == resource_module.RLIMIT_CPU

        # Check file size limit call
        fs_call = calls[1]
        assert fs_call[0][0] == resource_module.RLIMIT_FSIZE

        stack_call = calls[2]
        assert stack_call[0][0] == resource_module.RLIMIT_STACK

    @mock.patch('os.setpgid')
    @mock.patch('resource.setrlimit')
    @mock.patch('sys.platform', 'darwin')
    def test_preexec_fn_behavior_darwin(self, mock_setrlimit, mock_setpgid):
        """Test that preexec_fn actually calls the expected system functions."""
        import resource as resource_module

        params = ProgramParams(time_limit=5.0, fs_limit=1000, pgid=12345)

        preexec_fn = get_preexec_fn(params)
        preexec_fn()

        # Verify setpgid was called
        mock_setpgid.assert_called_once_with(0, 12345)

        # Verify setrlimit was called 2 times.
        assert mock_setrlimit.call_count == 2
        calls = mock_setrlimit.call_args_list

        # Check CPU limit call
        cpu_call = calls[0]
        assert cpu_call[0][0] == resource_module.RLIMIT_CPU

        # Check file size limit call
        fs_call = calls[1]
        assert fs_call[0][0] == resource_module.RLIMIT_FSIZE

    @mock.patch('psutil.Process')
    def test_handle_alarm_cpu_limit(self, mock_process_class):
        """Test _handle_alarm correctly detects CPU time limit violations."""
        mock_process = mock.MagicMock()
        mock_process_class.return_value = mock_process

        # Mock CPU times exceeding limit
        mock_process.cpu_times.return_value.user = 3.0
        mock_process.cpu_times.return_value.system = 2.5

        command = [sys.executable, '-c', 'import time; time.sleep(1)']
        params = ProgramParams(time_limit=5.0)  # CPU total will be 5.5 > 5.0

        program = Program(command, params)

        # Simulate the alarm handler
        program.params.time_limit = 5.0
        times = mock_process.cpu_times.return_value
        cpu_time = times.user + times.system

        # Verify the logic would trigger
        assert cpu_time > program.params.time_limit

        program.wait()  # Clean up

    @mock.patch('psutil.Process')
    def test_handle_alarm_memory_limit(self, mock_process_class):
        """Test _handle_alarm correctly detects memory limit violations."""
        mock_process = mock.MagicMock()
        mock_process_class.return_value = mock_process

        # Mock memory usage exceeding limit (20MB vs 10MB limit)
        mock_process.memory_info.return_value.rss = 20 * 1024 * 1024

        command = [sys.executable, '-c', 'import time; time.sleep(1)']
        params = ProgramParams(memory_limit=10)  # 10 MB limit

        program = Program(command, params)

        # Simulate the alarm handler logic
        memory_used = mock_process.memory_info.return_value.rss
        if program.params.memory_limit is not None:
            memory_limit_bytes = program.params.memory_limit * 1024 * 1024

            # Verify the logic would trigger
            assert memory_used > memory_limit_bytes

        program.wait()  # Clean up


class TestEdgeCases:
    """Test edge cases and error conditions."""

    def test_nonexistent_program(self):
        """Test execution of non-existent program."""
        params = ProgramParams()
        command = ['/nonexistent/program']

        with pytest.raises(ProgramError):
            Program(command, params)

    def test_empty_command(self):
        """Test execution with empty command."""
        params = ProgramParams()
        command = []

        with pytest.raises((ValueError, IndexError)):
            Program(command, params)

    def test_concurrent_programs(self, simple_hello_program):
        """Test running multiple programs concurrently."""
        params = ProgramParams()
        command = [sys.executable, str(simple_hello_program)]

        programs = []
        for _ in range(3):
            program = Program(command, params)
            programs.append(program)

        results = []
        for program in programs:
            result = program.wait()
            results.append(result)

        # All should succeed
        for result in results:
            assert result.exitcode == 0



import pathlib
import sys
from unittest.mock import patch

import pytest

from rbx.grading import steps
from rbx.grading.judge.sandbox import SandboxBase, SandboxParams
from rbx.grading.steps import (
    CoordinatedRunParams,
    GradingArtifacts,
    GradingFileInput,
    GradingFileOutput,
    GradingLogsHolder,
    RunLogMetadata,
)


@pytest.fixture(autouse=True)
def clear_steps_cache():
    """Clear all cached functions in steps.py before each test."""
    # Clear all cached functions to ensure clean state between tests
    steps._complain_about_clang.cache_clear()  # noqa: SLF001
    steps._get_cxx_version_output.cache_clear()  # noqa: SLF001
    steps._maybe_complain_about_sanitization.cache_clear()  # noqa: SLF001
    steps._try_following_alias_for_exe.cache_clear()  # noqa: SLF001


class TestStepsRunCoordinated:
    """Test the steps.run_coordinated function."""

    async def test_run_coordinated_simple_communication(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test basic communication between interactor and solution."""
        # Setup input files
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('interactor.stderr'),
                    dest=pathlib.Path('interactor.stderr'),
                ),
                GradingFileOutput(
                    src=pathlib.Path('solution.stderr'),
                    dest=pathlib.Path('solution.stderr'),
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        # Setup run parameters
        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(
                stderr_file=pathlib.Path('interactor.stderr'),
                timeout=5000,  # 5 seconds
            ),
            metadata=RunLogMetadata(language='python'),
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(
                stderr_file=pathlib.Path('solution.stderr'),
                timeout=5000,  # 5 seconds
            ),
            metadata=RunLogMetadata(language='python'),
        )

        # Run coordinated execution
        solution_log, interactor_log = await steps.run_coordinated(
            interactor_params, solution_params, artifacts, sandbox
        )

        # Verify results
        assert solution_log is not None
        assert interactor_log is not None
        assert solution_log.exitcode == 0
        assert interactor_log.exitcode == 0
        assert solution_log.time is not None
        assert interactor_log.time is not None
        assert solution_log.memory is not None
        assert interactor_log.memory is not None

        # Check that logs were captured
        assert artifacts.logs.run is not None
        assert artifacts.logs.interactor_run is not None
        assert artifacts.logs.run.exitcode == 0
        assert artifacts.logs.interactor_run.exitcode == 0

        # Check metadata was preserved
        assert solution_log.metadata is not None
        assert solution_log.metadata.language == 'python'
        assert interactor_log.metadata is not None
        assert interactor_log.metadata.language == 'python'

        # Check stderr files were created and verify content shows proper execution
        assert (cleandir / 'solution.stderr').exists()
        assert (cleandir / 'interactor.stderr').exists()

        # Read and verify stderr content to ensure programs executed correctly
        solution_stderr = (cleandir / 'solution.stderr').read_text()
        interactor_stderr = (cleandir / 'interactor.stderr').read_text()

        # Verify that the solution received responses (it logs these to stderr)
        assert 'Received: world' in solution_stderr
        assert 'Solution finished' in solution_stderr

        # Verify that the interactor also logged to stderr
        assert 'Interactor starting with 3 queries' in interactor_stderr
        assert 'Interactor finished' in interactor_stderr

    async def test_run_coordinated_with_timeout(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test run_coordinated with CPU timeout scenario."""
        interactor_file = testdata_path / 'steps_run_test' / 'busy_loop_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(timeout=100),  # 100ms CPU timeout
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(timeout=100),  # 100ms CPU timeout
        )

        solution_log, interactor_log = await steps.run_coordinated(
            interactor_params, solution_params, artifacts, sandbox
        )

        # Both logs should be present
        assert solution_log is not None
        assert interactor_log is not None

        # Check for timeout status - at least one should be affected by the busy loop
        timeout_statuses = [
            SandboxBase.EXIT_TIMEOUT,
        ]

        timeout_occurred = interactor_log.exitstatus in timeout_statuses

        assert timeout_occurred, (
            f'Expected CPU timeout to occur. '
            f'Solution: {solution_log.exitstatus}, Interactor: {interactor_log.exitstatus}'
        )
        assert interactor_log.exitindex < solution_log.exitindex

    async def test_run_coordinated_with_wall_timeout(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test run_coordinated with wall clock timeout scenario."""
        interactor_file = testdata_path / 'steps_run_test' / 'timeout_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(wallclock_timeout=2000),  # 2 second wall timeout
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(wallclock_timeout=2000),  # 2 second wall timeout
        )

        solution_log, interactor_log = await steps.run_coordinated(
            interactor_params, solution_params, artifacts, sandbox
        )

        # Both logs should be present
        assert solution_log is not None
        assert interactor_log is not None

        # Check for wall timeout status - at least one should be affected by the sleep
        wall_timeout_statuses = [
            SandboxBase.EXIT_TIMEOUT_WALL,
        ]

        wall_timeout_occurred = (
            interactor_log.exitstatus in wall_timeout_statuses
            or solution_log.exitstatus in wall_timeout_statuses
        )

        assert wall_timeout_occurred, (
            f'Expected wall timeout to occur. '
            f'Solution: {solution_log.exitstatus}, Interactor: {interactor_log.exitstatus}'
        )

    async def test_run_coordinated_solution_error(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test run_coordinated when solution crashes."""
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'error_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(timeout=5000),
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(timeout=5000),
        )

        solution_log, interactor_log = await steps.run_coordinated(
            interactor_params, solution_params, artifacts, sandbox
        )

        assert solution_log is not None
        assert interactor_log is not None

        # Solution should have non-zero exit code
        assert solution_log.exitcode != 0
        assert solution_log.exitstatus == SandboxBase.EXIT_NONZERO_RETURN
        assert solution_log.exitindex < interactor_log.exitindex

    async def test_run_coordinated_interactor_error(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test run_coordinated when interactor crashes."""
        interactor_file = testdata_path / 'steps_run_test' / 'error_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(timeout=5000),
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(timeout=5000),
        )

        solution_log, interactor_log = await steps.run_coordinated(
            interactor_params, solution_params, artifacts, sandbox
        )

        assert solution_log is not None
        assert interactor_log is not None

        # Interactor should have non-zero exit code
        assert interactor_log.exitcode != 0
        assert interactor_log.exitstatus == SandboxBase.EXIT_NONZERO_RETURN
        assert interactor_log.exitindex < solution_log.exitindex

    async def test_run_coordinated_java_memory_handling(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that Java commands have memory constraints removed."""
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        java_file = testdata_path / 'steps_run_test' / 'java_solution.java'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(
                    src=java_file, dest=pathlib.Path('java_solution.java')
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(
                timeout=5000,
                address_space=512,  # 512MB limit
            ),
        )

        # Java solution with memory constraints that should be removed
        solution_params = CoordinatedRunParams(
            command='javac java_solution.java && java java_solution',
            params=SandboxParams(
                timeout=5000,
                address_space=256,  # This should be removed for Java
            ),
        )

        # Mock the sandbox.run_communication method to verify the parameters
        with patch.object(
            sandbox, 'run_communication', wraps=sandbox.run_communication
        ) as mock_run_communication:
            solution_log, interactor_log = await steps.run_coordinated(
                interactor_params, solution_params, artifacts, sandbox
            )

            # Verify that run_communication was called
            mock_run_communication.assert_called_once()

            # Get the actual call arguments
            call_args = mock_run_communication.call_args

            # Arguments should be: solution_cmd, solution_params, interactor_cmd, interactor_params, merged_capture
            (
                solution_cmd,
                solution_params_actual,
                interactor_cmd,
                interactor_params_actual,
                merged_capture,
            ) = call_args[0]

            # Verify that the Java solution's address_space was removed
            assert solution_params_actual.address_space is None, (
                'Java solution should have address_space removed'
            )

            # Verify that the interactor still has its address_space constraint
            assert interactor_params_actual.address_space == 512, (
                'Interactor should keep its address_space constraint'
            )

        # The key test is that the function doesn't fail due to memory constraints
        # and that both logs are returned (even if the execution has issues)
        assert solution_log is not None
        assert interactor_log is not None

    async def test_run_coordinated_with_merged_capture(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test run_coordinated with merged capture file."""
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('merged.log'), dest=pathlib.Path('merged.log')
            )
        )
        artifacts.logs = GradingLogsHolder()

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(timeout=5000),
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(timeout=5000),
        )

        solution_log, interactor_log = await steps.run_coordinated(
            interactor_params,
            solution_params,
            artifacts,
            sandbox,
            merged_capture=pathlib.Path('merged.log'),
        )

        assert solution_log is not None
        assert interactor_log is not None

        # Check that merged capture file was created and verify its exact content
        assert (cleandir / 'merged.log').exists()

        # Read merged capture content
        merged_content = (cleandir / 'merged.log').read_text()

        # With the updated scripts that write in alternating turns, the output is now deterministic:
        # - Initial tee markers: < and >
        # - Solution outputs number of queries: >3
        # - Turn-based communication: >hello <world >test <echo_test >goodbye <echo_goodbye
        expected_content = """<
>
>3
>hello
<world
>test
<echo_test
>goodbye
<echo_goodbye
"""

        assert merged_content == expected_content

    async def test_run_coordinated_output_artifacts_failure(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test run_coordinated when output artifact processing fails."""
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        # Add a required output that won't be created
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('nonexistent.txt'),
                dest=pathlib.Path('nonexistent.txt'),
                optional=False,
            )
        )
        artifacts.logs = GradingLogsHolder()

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(timeout=5000),
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(timeout=5000),
        )

        result = await steps.run_coordinated(
            interactor_params, solution_params, artifacts, sandbox
        )

        # Should return (None, None) when output processing fails
        assert result == (None, None)

    async def test_run_coordinated_command_expansion(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that commands are properly split and expanded."""
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        # Use command with arguments that need to be split
        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} -u interactor.py',  # -u for unbuffered output
            params=SandboxParams(timeout=5000),
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} -u solution.py',  # -u for unbuffered output
            params=SandboxParams(timeout=5000),
        )

        solution_log, interactor_log = await steps.run_coordinated(
            interactor_params, solution_params, artifacts, sandbox
        )

        assert solution_log is not None
        assert interactor_log is not None
        assert solution_log.exitcode == 0
        assert interactor_log.exitcode == 0

    async def test_run_coordinated_handles_program_error(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """ProgramError from run_communication should map to sandbox error RunLogs."""
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(timeout=5000),
        )
        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py', params=SandboxParams(timeout=5000)
        )

        from rbx.grading.judge.program import ProgramError

        with patch.object(
            sandbox, 'run_communication', side_effect=ProgramError('boom')
        ):
            solution_log, interactor_log = await steps.run_coordinated(
                interactor_params, solution_params, artifacts, sandbox
            )

        assert solution_log is not None and interactor_log is not None
        assert solution_log.exitstatus == SandboxBase.EXIT_SANDBOX_ERROR
        assert interactor_log.exitstatus == SandboxBase.EXIT_SANDBOX_ERROR
        assert solution_log.exitcode == 1 and interactor_log.exitcode == 1

    async def test_run_coordinated_sandbox_reset_called(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that sandbox.reset() is called at the beginning."""
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(timeout=5000),
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(timeout=5000),
        )

        # Use mock.patch to track reset calls
        with patch.object(sandbox, 'reset', wraps=sandbox.reset) as mock_reset:
            await steps.run_coordinated(
                interactor_params, solution_params, artifacts, sandbox
            )

            # Verify reset was called exactly once
            mock_reset.assert_called_once()

    async def test_run_coordinated_without_logs_holder(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test run_coordinated without logs holder (should not crash)."""
        interactor_file = testdata_path / 'steps_run_test' / 'simple_interactor.py'
        solution_file = testdata_path / 'steps_run_test' / 'simple_solution.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(
                    src=interactor_file, dest=pathlib.Path('interactor.py')
                ),
                GradingFileInput(src=solution_file, dest=pathlib.Path('solution.py')),
            ]
        )
        # Note: No logs holder set

        interactor_params = CoordinatedRunParams(
            command=f'{sys.executable} interactor.py',
            params=SandboxParams(timeout=5000),
        )

        solution_params = CoordinatedRunParams(
            command=f'{sys.executable} solution.py',
            params=SandboxParams(timeout=5000),
        )

        solution_log, interactor_log = await steps.run_coordinated(
            interactor_params, solution_params, artifacts, sandbox
        )

        assert solution_log is not None
        assert interactor_log is not None
        assert solution_log.exitcode == 0
        assert interactor_log.exitcode == 0



import pathlib
from typing import Optional
from unittest.mock import AsyncMock, MagicMock

import pytest

from rbx.box import testcase_sample_utils
from rbx.box.generation_schema import GenerationMetadata, GenerationTestcaseEntry
from rbx.box.schema import Testcase
from rbx.box.testcase_utils import TestcaseEntry


@pytest.fixture
def mock_extract_generation_testcases_from_groups(monkeypatch):
    mock = AsyncMock()
    monkeypatch.setattr(
        'rbx.box.testcase_extractors.extract_generation_testcases_from_groups', mock
    )
    return mock


def create_entry(
    group_entry: TestcaseEntry,
    copied_to_input: pathlib.Path,
    copied_to_output: Optional[pathlib.Path] = None,
    copied_from_input: Optional[pathlib.Path] = None,
    copied_from_output: Optional[pathlib.Path] = None,
) -> GenerationTestcaseEntry:
    # Resolve all paths to avoid symlink issues
    copied_to_input = copied_to_input.resolve()
    if copied_to_output:
        copied_to_output = copied_to_output.resolve()
    if copied_from_input:
        copied_from_input = copied_from_input.resolve()
    if copied_from_output:
        copied_from_output = copied_from_output.resolve()

    copied_to = Testcase(inputPath=copied_to_input, outputPath=copied_to_output)

    copied_from = None
    if copied_from_input is not None:
        copied_from = Testcase(
            inputPath=copied_from_input, outputPath=copied_from_output
        )

    metadata = GenerationMetadata(
        copied_to=copied_to,
        copied_from=copied_from,
    )

    return GenerationTestcaseEntry(
        group_entry=group_entry,
        subgroup_entry=group_entry,  # Reuse for simplicity
        metadata=metadata,
    )


@pytest.mark.asyncio
async def test_get_statement_samples_basic_generated(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 1: Basic Generated - copied_to has input/output."""
    input_path = tmp_path / 'gen.in'
    output_path = tmp_path / 'gen.out'
    input_path.touch()
    output_path.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=input_path,
        copied_to_output=output_path,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()

    assert len(samples) == 1
    sample = samples[0]
    assert sample.inputPath.resolve() == input_path.resolve()
    assert sample.outputPath.resolve() == output_path.resolve()
    assert sample.answerPath.resolve() == output_path.resolve()
    assert sample.hasOutput is True
    assert sample.interaction is None


@pytest.mark.asyncio
async def test_get_statement_samples_basic_manual(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 2: Basic Manual - copied_from has input/output."""
    # copied_to paths (destination)
    dest_input = tmp_path / 'dest.in'
    dest_output = tmp_path / 'dest.out'
    # copied_from paths (source manual files)
    src_input = tmp_path / 'manual.in'
    src_output = tmp_path / 'manual.out'
    src_input.touch()
    src_output.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_to_output=dest_output,
        copied_from_input=src_input,
        copied_from_output=src_output,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()

    assert len(samples) == 1
    sample = samples[0]
    # Code behavior: unconditionally sets input_path to copied_to.inputPath
    assert sample.inputPath.resolve() == dest_input.resolve()
    # Code behavior: output_path initially src_output, NOT overwritten because dest_output missing
    assert sample.outputPath.resolve() == src_output.resolve()
    assert sample.answerPath.resolve() == src_output.resolve()
    assert sample.hasOutput is True


@pytest.mark.asyncio
async def test_get_statement_samples_pin_file_generated(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 5: Pin File (Generated) - .pin exists next to copied_to."""
    dest_input = tmp_path / 'dest.in'
    dest_pin = tmp_path / 'dest.pin'
    dest_pin.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    assert list(samples)[0].inputPath.resolve() == dest_pin.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_pout_file_generated(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 6: Pout File (Generated) - .pout exists next to copied_to."""
    dest_input = tmp_path / 'dest.in'
    dest_pout = tmp_path / 'dest.pout'
    dest_pout.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    assert list(samples)[0].outputPath.resolve() == dest_pout.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_pin_file_manual(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 7: Pin File (Manual) - .pin exists next to copied_from."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_pin = tmp_path / 'manual.pin'
    src_pin.touch()

    # Even if generated exists, manual pin should win because process_additional_files(manual) is last
    dest_input.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    assert list(samples)[0].inputPath.resolve() == src_pin.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_pout_file_manual(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 8: Pout File (Manual) - .pout exists next to copied_from."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_pout = tmp_path / 'manual.pout'
    src_pout.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    assert list(samples)[0].outputPath.resolve() == src_pout.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_explanation_manual(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 9: Explanation (Manual) - explanation_suffix provided."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_expl = tmp_path / 'manual.desc'
    src_expl.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples(
        explanation_suffix='.desc'
    )
    assert list(samples)[0].explanationPath.resolve() == src_expl.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_explanation_priority(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 10: Explanation Priority - Manual vs Generated. Manual should win."""
    dest_input = tmp_path / 'dest.in'
    dest_expl = tmp_path / 'dest.desc'
    dest_expl.touch()

    src_input = tmp_path / 'manual.in'
    src_expl = tmp_path / 'manual.desc'
    src_expl.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples(
        explanation_suffix='.desc'
    )
    assert list(samples)[0].explanationPath.resolve() == src_expl.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_out_check_manual(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 11: .out check (Manual)."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_out = tmp_path / 'manual.out'
    src_out.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    assert list(samples)[0].outputPath.resolve() == src_out.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_out_statement_check_manual(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 12: .out.statement check (Manual)."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_out_stmt = tmp_path / 'manual.out.statement'
    src_out_stmt.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    assert list(samples)[0].outputPath.resolve() == src_out_stmt.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_out_statement_vs_out(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 13: .out.statement vs .out. .out.statement should win."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_out = tmp_path / 'manual.out'
    src_out.touch()
    src_out_stmt = tmp_path / 'manual.out.statement'
    src_out_stmt.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    assert list(samples)[0].outputPath.resolve() == src_out_stmt.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_interaction_generated(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 14: Interaction (Generated)."""
    dest_input = tmp_path / 'dest.in'
    dest_intr = tmp_path / 'dest.interaction'
    dest_intr.write_text('> 5\n< 10')  # Basic interaction format

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]
    assert sample.interaction is not None
    # We aren't testing parse_interaction depth here, just that it was picked up


@pytest.mark.asyncio
async def test_get_statement_samples_interaction_manual(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 15: Interaction (Manual)."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_intr = tmp_path / 'manual.interaction'
    src_intr.write_text('> 5\n< 10')

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    # Code analysis reveals that manual interaction is looked up relative to 'input_path'.
    # Since input_path is overwritten by dest_input, and dest_input has no interaction,
    # this returns None.
    # To make this pass with current code, we expect None.
    # This highlights a potential issue/feature: Manual interactions are ignored if dest input is used.
    assert list(samples)[0].interaction is None


@pytest.mark.asyncio
async def test_get_statement_samples_empty_sentinel(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 16: Empty Sentinel - No files exist."""
    dest_input = tmp_path / 'dest.in'  # Doesn't exist

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]
    # Should point to sentinel?
    # Actually copied_to.inputPath is passed directly if no overrides
    assert sample.inputPath.resolve() == dest_input.resolve()

    # Let's verify it actually is the sentinel or similar valid path
    from rbx import utils

    assert sample.outputPath.resolve() == utils.get_empty_sentinel_path().resolve()
    assert sample.hasOutput is True


@pytest.mark.asyncio
async def test_get_statement_samples_no_output_file_but_path_set(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 17: Output path set but file missing."""
    dest_input = tmp_path / 'dest.in'
    dest_output = tmp_path / 'dest.out'
    # Don't create dest_output file

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_to_output=dest_output,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]

    from rbx import utils

    assert sample.outputPath.resolve() == utils.get_empty_sentinel_path().resolve()
    assert sample.hasOutput is True


@pytest.mark.asyncio
async def test_get_statement_samples_answer_path_stability(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 18: Answer Path stability."""
    dest_input = tmp_path / 'dest.in'
    dest_output = tmp_path / 'dest.out'
    dest_output.touch()

    # Add a .pout file
    dest_pout = tmp_path / 'dest.pout'
    dest_pout.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_to_output=dest_output,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]

    # Logic:
    # 1. output_path = dest_output, answer_path = dest_output
    # 2. process_additional_files -> finds .pout -> output_path = dest_pout
    # answer_path is NOT updated in process_additional_files

    assert sample.outputPath.resolve() == dest_pout.resolve()
    assert sample.answerPath.resolve() == dest_output.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_broken_interaction(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 19: Broken Interaction - Verify error/exit."""
    dest_input = tmp_path / 'dest.in'
    dest_intr = tmp_path / 'dest.interaction'
    dest_intr.write_text('BROKEN FILE FORMAT')

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    with pytest.raises(testcase_sample_utils.typer.Exit):
        await testcase_sample_utils.get_statement_samples()


@pytest.mark.asyncio
async def test_get_statement_samples_sentinel_resolution(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 20: Ensure paths are resolved to absolute."""
    # Using relative paths (if poss) or just checking the result is absolute
    dest_input = tmp_path / 'dest.in'
    dest_input.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]

    assert sample.inputPath.is_absolute()
    assert sample.outputPath.is_absolute()


@pytest.mark.asyncio
async def test_get_statement_samples_check_output_true(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 21: checkOutput=True - Manual .out override different from answers."""
    dest_input = tmp_path / 'dest.in'
    dest_output = tmp_path / 'dest.out'
    dest_output.touch()

    src_input = tmp_path / 'manual.in'
    src_out = tmp_path / 'manual.out'
    src_out.write_text('manual output')

    # copied_to has output (so answer_path = dest_output)
    # manual .out exists (so output_path = src_out via process_additional_files on copied_from)

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_to_output=dest_output,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]

    assert sample.hasOutput is True
    assert sample.checkOutput is True
    assert sample.outputPath.resolve() == src_out.resolve()
    assert sample.answerPath.resolve() == dest_output.resolve()


@pytest.mark.asyncio
async def test_get_statement_samples_check_output_false_same_file(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 22: checkOutput=False - Output path equals answer path."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_out = tmp_path / 'manual.out'
    src_out.touch()

    # No dest_output, so answer_path comes from manual if copied_from has it.
    # copied_from logic:
    # if copied_from.outputPath exists: output_path=it, answer_path=it.
    # checking create_entry: it sets copied_from.outputPath if arg provided.

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
        copied_from_output=src_out,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]

    assert sample.outputPath.resolve() == src_out.resolve()
    assert sample.answerPath.resolve() == src_out.resolve()
    assert sample.checkOutput is False


@pytest.mark.asyncio
async def test_get_statement_samples_check_output_false_statement_file(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 23: checkOutput=False - .out.statement file usage."""
    dest_input = tmp_path / 'dest.in'
    dest_output = tmp_path / 'dest.out'
    dest_output.touch()

    src_input = tmp_path / 'manual.in'
    src_stmt = tmp_path / 'manual.out.statement'
    src_stmt.touch()

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_to_output=dest_output,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]

    assert sample.outputPath.resolve() == src_stmt.resolve()
    assert sample.checkOutput is False


@pytest.mark.asyncio
async def test_get_statement_samples_check_output_no_answer(
    tmp_path, mock_extract_generation_testcases_from_groups
):
    """Test 24: checkOutput=False - No answer path available."""
    dest_input = tmp_path / 'dest.in'

    src_input = tmp_path / 'manual.in'
    src_out = tmp_path / 'manual.out'
    src_out.touch()

    # copied_to_output=None, copied_from_output=None
    # But manual .out exists, detected by process_additional_files

    entry = create_entry(
        TestcaseEntry(group='samples', index=0),
        copied_to_input=dest_input,
        copied_from_input=src_input,
    )
    mock_extract_generation_testcases_from_groups.return_value = [entry]

    samples = await testcase_sample_utils.get_statement_samples()
    sample = list(samples)[0]

    assert sample.outputPath.resolve() == src_out.resolve()
    assert sample.answerPath is None
    assert sample.checkOutput is False


@pytest.fixture
def mock_dependencies_for_build_samples(monkeypatch):
    """Mocks all external dependencies for build_samples."""
    mocks = {
        'build': AsyncMock(return_value=True),
        'get_statement_samples': AsyncMock(),
        'compile_output_validators': MagicMock(),
        'validate_file': AsyncMock(),
        'print_validation_report': MagicMock(),
        'compile_checker': AsyncMock(return_value='checker_digest'),
        'check_sample': AsyncMock(return_value=True),
        'console_print': MagicMock(),
    }

    monkeypatch.setattr('rbx.box.builder.build', mocks['build'])
    monkeypatch.setattr(
        'rbx.box.testcase_sample_utils.get_statement_samples',
        mocks['get_statement_samples'],
    )
    monkeypatch.setattr(
        'rbx.box.testcase_sample_utils.compile_output_validators_for_entries',
        mocks['compile_output_validators'],
    )
    monkeypatch.setattr('rbx.box.validators.validate_file', mocks['validate_file'])
    monkeypatch.setattr(
        'rbx.box.validators.print_validation_report', mocks['print_validation_report']
    )
    monkeypatch.setattr('rbx.box.checkers.compile_checker', mocks['compile_checker'])
    monkeypatch.setattr(
        'rbx.box.testcase_sample_utils._check_sample', mocks['check_sample']
    )
    monkeypatch.setattr('rbx.console.console.print', mocks['console_print'])

    # Patch TestcaseValidationInfo to satisfy Pydantic validation with mocks
    class MockValidationInfo:
        def __init__(self, **kwargs):
            self.ok = kwargs.get('ok')
            for k, v in kwargs.items():
                setattr(self, k, v)

    monkeypatch.setattr(
        'rbx.box.testcase_sample_utils.TestcaseValidationInfo', MockValidationInfo
    )

    return mocks


@pytest.mark.asyncio
async def test_build_samples_success(
    mock_dependencies_for_build_samples,
):
    """Test 25: build_samples success path (validation + checker pass)."""
    mocks = mock_dependencies_for_build_samples

    # Setup samples with checkOutput=True
    sample = AsyncMock()
    sample.checkOutput = True
    sample.entry = AsyncMock()
    val_mock = MagicMock()
    val_mock.path = 'val1'
    sample.entry.output_validators = [val_mock]
    sample.entry.group_entry.group = 'samples'
    sample.entry.metadata = 'metadata'
    sample.outputPath = 'path'
    mocks['get_statement_samples'].return_value = [sample]

    # Setup validation success
    mocks['compile_output_validators'].return_value = {'val1': 'digest'}
    mocks['validate_file'].return_value = (True, 'ok', {})

    result = await testcase_sample_utils.build_samples(verification=10, validate=True)

    assert result is True

    # Verify validation flow
    mocks['compile_output_validators'].assert_called_once()
    mocks['validate_file'].assert_called_once()
    mocks['print_validation_report'].assert_called_once()

    # Verify checker flow
    mocks['compile_checker'].assert_called_once()
    mocks['check_sample'].assert_called_once()


@pytest.mark.asyncio
async def test_build_samples_validation_fail(
    mock_dependencies_for_build_samples,
):
    """Test 26: build_samples validation fails -> checker skipped."""
    mocks = mock_dependencies_for_build_samples

    sample = AsyncMock()
    sample.checkOutput = True
    sample.entry = AsyncMock()
    val_mock = MagicMock()
    val_mock.path = 'val1'
    sample.entry.output_validators = [val_mock]
    sample.entry.group_entry.group = 'samples'
    sample.entry.metadata = 'metadata'
    sample.outputPath = 'path'
    mocks['get_statement_samples'].return_value = [sample]

    # Validation fails
    mocks['compile_output_validators'].return_value = {'val1': 'digest'}
    mocks['validate_file'].return_value = (False, 'fail', {})

    result = await testcase_sample_utils.build_samples(verification=10, validate=True)

    assert result is False

    # Check that checker was NOT called
    mocks['compile_checker'].assert_not_called()
    mocks['check_sample'].assert_not_called()


@pytest.mark.asyncio
async def test_build_samples_checker_fail(
    mock_dependencies_for_build_samples,
):
    """Test 27: build_samples validation passes -> checker fails."""
    mocks = mock_dependencies_for_build_samples

    sample = AsyncMock()
    sample.checkOutput = True
    sample.entry = AsyncMock()
    val_mock = MagicMock()
    val_mock.path = 'val1'
    sample.entry.output_validators = [val_mock]
    sample.entry.group_entry.group = 'samples'
    sample.entry.metadata = 'metadata'
    sample.outputPath = 'path'
    mocks['get_statement_samples'].return_value = [sample]

    # Validation passes
    mocks['validate_file'].return_value = (True, 'ok', {})

    # Checker fails
    mocks['check_sample'].return_value = False

    result = await testcase_sample_utils.build_samples(verification=10, validate=True)

    assert result is False

    mocks['validate_file'].assert_called()
    mocks['compile_checker'].assert_called_once()
    mocks['check_sample'].assert_called_once()


@pytest.mark.asyncio
async def test_build_samples_no_validate_flag(
    mock_dependencies_for_build_samples,
):
    """Test 28: build_samples with validate=False -> skip all."""
    mocks = mock_dependencies_for_build_samples

    result = await testcase_sample_utils.build_samples(verification=10, validate=False)

    assert result is True

    # Should skip everything after builder.build
    mocks['get_statement_samples'].assert_not_called()
    mocks['validate_file'].assert_not_called()
    mocks['check_sample'].assert_not_called()


@pytest.mark.asyncio
async def test_build_samples_builder_fails(
    mock_dependencies_for_build_samples,
):
    """Test 29: build_samples builder fails -> fail immediately."""
    mocks = mock_dependencies_for_build_samples
    mocks['build'].return_value = False

    result = await testcase_sample_utils.build_samples(verification=10, validate=True)

    assert result is False
    mocks['get_statement_samples'].assert_not_called()


@pytest.mark.asyncio
async def test_build_samples_check_outputs_only_false(
    mock_dependencies_for_build_samples,
):
    """Test 30: build_samples with check_outputs_only=False -> builder called."""
    mocks = mock_dependencies_for_build_samples

    result = await testcase_sample_utils.build_samples(
        verification=10, validate=True, check_outputs_only=False
    )

    assert result is True
    mocks['build'].assert_called_once()


@pytest.mark.asyncio
async def test_build_samples_check_outputs_only_true(
    mock_dependencies_for_build_samples,
):
    """Test 31: build_samples with check_outputs_only=True -> builder NOT called."""
    mocks = mock_dependencies_for_build_samples

    result = await testcase_sample_utils.build_samples(
        verification=10, validate=True, check_outputs_only=True
    )

    assert result is True
    mocks['build'].assert_not_called()



from __future__ import annotations

import importlib.resources
import logging
import os
import pathlib
import shutil
import signal
import subprocess
import sys
import tempfile
import typing
from typing import List, Optional, Tuple

from rbx import utils
from rbx.grading.judge.cacher import FileCacher
from rbx.grading.judge.program import (
    FileLike,
    Program,
    ProgramCode,
    ProgramIO,
    ProgramParams,
    ProgramResult,
)
from rbx.grading.judge.sandbox import (
    CommunicationParams,
    SandboxBase,
    SandboxLog,
    SandboxParams,
)

logger = logging.getLogger(__name__)

TEE_CODE = R"""
import sys
c = sys.argv[1]
new = True
while True:
    l = sys.stdin.read(1)
    if l=='': break
    sys.stdout.write(l)
    sys.stdout.flush()
    if new: sys.stderr.write(c)
    sys.stderr.write(l)
    sys.stderr.flush()
    new = l=='\n'
"""


class StupidSandbox(SandboxBase):
    """A stupid sandbox implementation. It has very few features and
    is not secure against things like box escaping and fork
    bombs. Yet, it is very portable and has no dependencies, so it's
    very useful for testing. Using in real contests is strongly
    discouraged.

    """

    exec_num: int

    def __init__(
        self,
        file_cacher: Optional[FileCacher] = None,
        name: Optional[str] = None,
        temp_dir: Optional[pathlib.Path] = None,
    ):
        """Initialization.

        For arguments documentation, see SandboxBase.__init__.

        """
        if not temp_dir:
            temp_dir = pathlib.Path(tempfile.gettempdir())
        SandboxBase.__init__(self, file_cacher, name, temp_dir)

        # Make box directory
        self.initialize()

    def initialize(self):
        self._path = pathlib.Path(
            tempfile.mkdtemp(dir=str(self.temp_dir), prefix='rbx-%s-' % (self.name))
        )
        self.exec_num = -1
        self.log = None
        self.returncode = None
        self._path.mkdir(parents=True, exist_ok=True)

        logger.debug("Sandbox in `%s' created, using stupid box.", self._path)

        # Box parameters
        self.chdir = self._path

    def get_root_path(self) -> pathlib.Path:
        """Return the toplevel path of the sandbox.

        return (Path): the root path.

        """
        return self._path

    def use_soft_timeout(self) -> bool:
        return True

    def _get_exit_status(self, result: ProgramResult) -> str:
        if ProgramCode.TE in result.program_codes:
            return SandboxBase.EXIT_TERMINATED
        if ProgramCode.WT in result.program_codes:
            return SandboxBase.EXIT_TIMEOUT_WALL
        if ProgramCode.TO in result.program_codes:
            return SandboxBase.EXIT_TIMEOUT
        if ProgramCode.OL in result.program_codes:
            return SandboxBase.EXIT_OUTPUT_LIMIT_EXCEEDED
        if ProgramCode.ML in result.program_codes:
            return SandboxBase.EXIT_MEMORY_LIMIT_EXCEEDED
        if ProgramCode.SG in result.program_codes:
            return SandboxBase.EXIT_SIGNAL
        if ProgramCode.RE in result.program_codes:
            return SandboxBase.EXIT_NONZERO_RETURN
        return SandboxBase.EXIT_OK

    def _get_io(self, params: SandboxParams, pipe_io: bool = False) -> ProgramIO:
        io = ProgramIO()
        if params.stdin_file and not pipe_io:
            io.input = self.relative_path(params.stdin_file)
        if params.stdout_file and not pipe_io:
            io.output = self.relative_path(params.stdout_file)
        if params.stderr_file:
            io.stderr = self.relative_path(params.stderr_file)
        return io

    def _get_program_params(self, params: SandboxParams) -> ProgramParams:
        return ProgramParams(
            chdir=self.chdir,
            time_limit=params.timeout / 1000 if params.timeout else None,
            wall_time_limit=params.wallclock_timeout / 1000
            if params.wallclock_timeout
            else None,
            memory_limit=params.address_space,
            fs_limit=params.fsize,
            env=params.set_env,
            io=self._get_io(params),
        )

    def _get_tee_program_params(self, io: ProgramIO, pgid: int) -> ProgramParams:
        return ProgramParams(
            chdir=self.chdir,
            time_limit=None,
            wall_time_limit=None,
            memory_limit=None,
            io=io,
            pgid=pgid,
        )

    def _get_sandbox_log(
        self, result: ProgramResult, params: SandboxParams
    ) -> SandboxLog:
        return SandboxLog(
            params=params.model_copy(deep=True),
            execution_time=result.cpu_time,
            wall_time=result.wall_time,
            memory_used=result.memory_used,
            exitcode=result.exitcode,
            exitstatus=self._get_exit_status(result),
            killing_signal=result.killing_signal,
            other_logs={
                'program_codes': [code.value for code in result.program_codes],
                'alarm_msg': result.alarm_msg,
                'wall_time': result.wall_time,
            },
        )

    def _needs_teeing(
        self,
        params: SandboxParams,
        interactor_params: SandboxParams,
        merged_capture: Optional[pathlib.Path] = None,
    ) -> bool:
        return (
            params.stdout_file is not None
            or interactor_params.stdout_file is not None
            or merged_capture is not None
        )

    def _get_tee_executable(
        self, communication_params: CommunicationParams
    ) -> pathlib.Path:
        if communication_params.tee_mode == 'line':
            with importlib.resources.as_file(
                importlib.resources.files('rbx')
                / 'grading'
                / 'judge'
                / 'sandboxes'
                / 'line_tee.py'
            ) as file:
                return file
        with importlib.resources.as_file(
            importlib.resources.files('rbx')
            / 'grading'
            / 'judge'
            / 'sandboxes'
            / 'tee.py'
        ) as file:
            return file

    def _get_tee_command(
        self,
        char: str,
        communication_params: CommunicationParams,
        extra: Optional[str] = None,
    ) -> List[str]:
        return [
            sys.executable,
            str(utils.abspath(self._get_tee_executable(communication_params))),
            char,
            extra or '/dev/null',
        ]

    def _get_tee_program(
        self,
        char: str,
        stdin: FileLike,
        stdout: FileLike,
        communication_params: CommunicationParams,
        pgid: int,
        capture: Optional[pathlib.Path] = None,
    ) -> Program:
        io = ProgramIO(input=stdin, output=stdout, stderr=subprocess.DEVNULL)
        if communication_params.merged_capture:
            io.stderr = self.relative_path(communication_params.merged_capture).open(
                'ab'
            )
        return Program(
            self._get_tee_command(
                char,
                communication_params,
                str(self.relative_path(capture)) if capture else None,
            ),
            self._get_tee_program_params(io, pgid),
        )

    def _get_pathlike_stdout(self, io: ProgramIO) -> Optional[pathlib.Path]:
        if isinstance(io.output, str) or isinstance(io.output, pathlib.Path):
            return pathlib.Path(io.output)
        return None

    def run(self, command: List[str], params: SandboxParams) -> SandboxLog:
        self.exec_num += 1

        logger.debug(
            "Executing program in sandbox with command: `%s'.", ' '.join(command)
        )
        with open(
            self.relative_path(self.cmd_file), 'at', encoding='utf-8'
        ) as commands:
            commands.write('%s\n' % command)

        program = Program(command, self._get_program_params(params))
        result = program.wait()

        return self._get_sandbox_log(result, params)

    def run_communication(
        self,
        command: List[str],
        params: SandboxParams,
        interactor_command: List[str],
        interactor_params: SandboxParams,
        communication_params: Optional[CommunicationParams] = None,
    ) -> Tuple[SandboxLog, SandboxLog]:
        communication_params = communication_params or CommunicationParams()
        self.exec_num += 1

        logger.debug(
            "Executing program in sandbox with command: `%s'.", ' '.join(command)
        )
        with open(
            self.relative_path(self.cmd_file), 'at', encoding='utf-8'
        ) as commands:
            commands.write('%s\n' % command)

        interactor_program_params = self._get_program_params(interactor_params)
        interactor_program_params.io = self._get_io(interactor_params, pipe_io=True)
        interactor = Program(
            interactor_command,
            interactor_program_params,
        )
        assert interactor.pipes.output is not None
        assert interactor.pipes.input is not None
        solution_input_pipe = interactor.pipes.output
        solution_output_pipe = interactor.pipes.input

        group_id = os.getpgid(interactor.pid)
        should_tee = self._needs_teeing(
            params, interactor_params, communication_params.merged_capture
        )

        if should_tee:
            if communication_params.merged_capture:
                self.create_file_from_string(
                    communication_params.merged_capture, '<\n>\n', override=True
                )

            solution_tee = self._get_tee_program(
                '>',
                stdin=subprocess.PIPE,
                stdout=interactor.pipes.input,
                capture=self._get_pathlike_stdout(self._get_io(params)),
                pgid=group_id,
                communication_params=communication_params,
            )
            interactor_tee = self._get_tee_program(
                '<',
                stdin=interactor.pipes.output,
                stdout=subprocess.PIPE,
                capture=self._get_pathlike_stdout(self._get_io(interactor_params)),
                communication_params=communication_params,
                pgid=group_id,
            )
            assert solution_tee.pipes.input is not None
            assert interactor_tee.pipes.output is not None
            solution_input_pipe = interactor_tee.pipes.output
            solution_output_pipe = solution_tee.pipes.input

        program_params = self._get_program_params(params)
        program_params.io = self._get_io(params, pipe_io=True)
        program_params.io.input = solution_input_pipe
        program_params.io.output = solution_output_pipe
        program_params.pgid = group_id
        program = Program(command, program_params)

        results: List[Optional[SandboxLog]] = [None, None]

        for idx in range(4 if should_tee else 2):
            pid, status, ru = os.wait4(-group_id, 0)

            if pid == interactor.pid:
                program_result = interactor.process_exit(status, ru)
                results[1] = self._get_sandbox_log(program_result, interactor_params)
                results[1].exit_index = idx

                interactor.pipes.output.close()
                if should_tee:
                    assert interactor_tee.pipes.output is not None
                    interactor_tee.pipes.output.close()

                if idx == 0 and program_result.exitcode != 0:
                    try:
                        os.killpg(group_id, signal.SIGKILL)
                    except Exception:
                        pass
            elif pid == program.pid:
                program_result = program.process_exit(status, ru)
                results[0] = self._get_sandbox_log(program_result, params)
                results[0].exit_index = idx

                interactor.pipes.input.close()
                if should_tee:
                    assert solution_tee.pipes.input is not None
                    solution_tee.pipes.input.close()
            elif should_tee and (pid in (solution_tee.pid, interactor_tee.pid)):
                pass
            else:
                raise RuntimeError(f'Unknown pid: {pid}')

        for p in (interactor, program):
            p.close()
        if should_tee:
            for p in (solution_tee, interactor_tee):
                p.close()

        return typing.cast(Tuple[SandboxLog, SandboxLog], tuple(results))

    def cleanup(self, delete=False):
        """See Sandbox.cleanup()."""
        # This sandbox doesn't have any cleanup, but we might want to delete.
        if delete:
            logger.debug('Deleting sandbox in %s.', self._path)
            shutil.rmtree(str(self._path), ignore_errors=True)



import pathlib

import typer

from rbx import annotations
from rbx.box.tooling.boca.scrape import scrape_boca
from rbx.box.tooling.boca.scraper import get_boca_scraper
from rbx.box.tooling.boca.ui import run_app

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


@app.command('scrape', help='Scrape runs from BOCA.')
def scrape() -> None:
    scrape_boca(pathlib.Path())


@app.command('view', help='Open Textual UI to visualize BOCA submissions.')
def view(
    contest_id: str = typer.Option(
        None,
        '--contest-id',
        '-c',
        prompt='Contest ID',
        help='Contest identifier to load (stored under app data).',
    ),
) -> None:
    # Normalize empty input to None to let the UI apply default
    cid = (contest_id or '').strip() or None
    run_app(contest_id=cid)


@app.command('submit', help='Submit solutions to BOCA.')
def submit() -> None:
    from rbx.box.tooling.boca.submitter import judge_all, submit_all_solutions

    judged_solutions = list(submit_all_solutions(get_boca_scraper(is_judge=True)))
    judge_all(judged_solutions)
